from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta
import hashlib
import json
import os
from pathlib import Path
import time
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd

try:  # pragma: no cover
    import torch.distributed as dist  # type: ignore
except Exception:  # pragma: no cover
    dist = None  # type: ignore

from ins_pricing.modelling.bayesopt.utils.distributed_utils import (
    DistributedUtils,
    free_cuda as distributed_free_cuda,
)


@dataclass(frozen=True)
class BayesOptRunnerDeps:
    ropt: Any
    pytorch_trainers: Sequence[str]
    build_model_names: Callable[[Any, Any], List[str]]
    dedupe_preserve_order: Callable[[List[str]], List[str]]
    load_dataset: Callable[..., Any]
    resolve_data_path: Callable[..., Path]
    resolve_path: Callable[..., Optional[Path]]
    fingerprint_file: Callable[..., Dict[str, Any]]
    coerce_dataset_types: Callable[[Any], Any]
    split_train_test: Callable[..., Tuple[Any, Any]]
    resolve_and_load_config: Callable[..., Tuple[Path, Dict[str, Any]]]
    resolve_data_config: Callable[..., Tuple[Any, Any, Any, Any]]
    resolve_report_config: Callable[[Dict[str, Any]], Dict[str, Any]]
    resolve_split_config: Callable[[Dict[str, Any]], Dict[str, Any]]
    resolve_runtime_config: Callable[[Dict[str, Any]], Dict[str, Any]]
    resolve_output_dirs: Callable[..., Dict[str, Any]]


@dataclass(frozen=True)
class BayesOptRunnerHooks:
    plot_loss_curve_for_trainer: Callable[[str, Any], None]
    compute_psi_report: Callable[..., Any]
    evaluate_and_report: Callable[..., None]
    plot_curves_for_model: Callable[[Any, List[str], Dict[str, Any]], None]


def _create_ddp_barrier(dist_ctx: Any):
    """Create a DDP barrier function for distributed training synchronization."""

    def _ddp_barrier(reason: str) -> None:
        if not getattr(dist_ctx, "is_distributed", False):
            return
        if dist is None:
            return
        try:
            if not getattr(dist, "is_available", lambda: False)():
                return
            if not dist.is_initialized():
                ddp_ok, _, _, _ = DistributedUtils.setup_ddp()
                if not ddp_ok or not dist.is_initialized():
                    return
        except Exception as exc:
            print(f"[DDP] barrier pre-check failed during {reason}: {exc}", flush=True)
            raise

        timeout_seconds = int(os.environ.get("BAYESOPT_DDP_BARRIER_TIMEOUT", "1800"))
        debug_barrier = os.environ.get("BAYESOPT_DDP_BARRIER_DEBUG", "").strip().lower() in {
            "1",
            "true",
            "yes",
            "on",
        }
        rank = None
        world = None
        if debug_barrier:
            try:
                rank = dist.get_rank()
                world = dist.get_world_size()
                print(
                    f"[DDP] entering barrier({reason}) rank={rank}/{world}",
                    flush=True,
                )
            except Exception:
                debug_barrier = False
        try:
            timeout = timedelta(seconds=max(1, timeout_seconds))
            backend = None
            try:
                backend = dist.get_backend()
            except Exception:
                backend = None

            monitored = getattr(dist, "monitored_barrier", None)
            if backend == "gloo" and callable(monitored):
                monitored(timeout=timeout)
            else:
                work = None
                try:
                    work = dist.barrier(async_op=True)
                except TypeError:
                    work = None
                if work is not None:
                    wait = getattr(work, "wait", None)
                    if callable(wait):
                        try:
                            wait(timeout=timeout)
                        except TypeError:
                            wait()
                    else:
                        dist.barrier()
                else:
                    dist.barrier()
            if debug_barrier:
                print(
                    f"[DDP] exit barrier({reason}) rank={rank}/{world}",
                    flush=True,
                )
        except Exception as exc:
            print(f"[DDP] barrier failed during {reason}: {exc}", flush=True)
            raise

    return _ddp_barrier


def _free_cuda_safe(ropt_module: Any) -> None:
    """Release CUDA memory using the best available compatibility path."""
    free_fn = getattr(ropt_module, "free_cuda", None)
    if callable(free_fn):
        free_fn()
        return
    distributed_free_cuda()


def _stream_random_split_csv(
    data_path: Path,
    *,
    holdout_ratio: float,
    rand_seed: Optional[int],
    dtype_map: Dict[str, Any],
    usecols: Optional[List[str]],
    low_memory: bool,
    chunksize: int,
    coerce_dataset_types: Callable[[Any], Any],
) -> Tuple[pd.DataFrame, pd.DataFrame, int]:
    """Stream CSV chunks and split rows by Bernoulli sampling to avoid full-file loading."""
    holdout_ratio = float(holdout_ratio)
    if not (0.0 < holdout_ratio < 1.0):
        raise ValueError(
            f"holdout_ratio must be in (0, 1) for streaming random split; got {holdout_ratio}."
    )
    chunk_size = max(1, int(chunksize))
    rng = np.random.default_rng(rand_seed if rand_seed is not None else 13)
    read_kwargs: Dict[str, Any] = {"low_memory": bool(low_memory), "chunksize": chunk_size}
    if usecols:
        read_kwargs["usecols"] = list(usecols)
    if dtype_map:
        dtype_selected = dict(dtype_map)
        if usecols:
            allowed = set(usecols)
            dtype_selected = {k: v for k, v in dtype_selected.items() if k in allowed}
        if dtype_selected:
            read_kwargs["dtype"] = dtype_selected

    train_parts: List[pd.DataFrame] = []
    test_parts: List[pd.DataFrame] = []
    total_rows = 0
    columns: Optional[pd.Index] = None

    for chunk in pd.read_csv(data_path, **read_kwargs):
        chunk = coerce_dataset_types(chunk)
        if columns is None:
            columns = chunk.columns
        n_chunk = int(len(chunk))
        if n_chunk == 0:
            continue
        total_rows += n_chunk
        mask_test = rng.random(n_chunk) < holdout_ratio
        if mask_test.any():
            test_parts.append(chunk.loc[mask_test])
        mask_train = ~mask_test
        if mask_train.any():
            train_parts.append(chunk.loc[mask_train])

    if total_rows == 0:
        raise ValueError(f"Dataset is empty: {data_path}")

    if columns is None:
        columns = pd.Index([])
    train_df = pd.concat(train_parts, ignore_index=True) if train_parts else pd.DataFrame(columns=columns)
    test_df = pd.concat(test_parts, ignore_index=True) if test_parts else pd.DataFrame(columns=columns)
    if train_df.empty or test_df.empty:
        raise ValueError(
            "Streaming split produced an empty train or test set. "
            "Adjust holdout_ratio or disable stream_split_csv."
        )
    return train_df, test_df, total_rows


def _dedupe_columns(columns: List[str]) -> List[str]:
    out: List[str] = []
    seen = set()
    for col in columns:
        if not isinstance(col, str):
            continue
        key = col.strip()
        if not key or key in seen:
            continue
        seen.add(key)
        out.append(key)
    return out


def _resolve_required_columns(
    cfg: Dict[str, Any],
    *,
    split_group_col: Optional[str],
    split_time_col: Optional[str],
    cv_group_col: Optional[str],
    cv_time_col: Optional[str],
    report_group_cols: Optional[List[str]],
    report_time_col: Optional[str],
) -> List[str]:
    cols: List[str] = []
    feature_list = cfg.get("feature_list") or []
    if isinstance(feature_list, (list, tuple)):
        cols.extend([c for c in feature_list if isinstance(c, str)])
    target = cfg.get("target")
    weight = cfg.get("weight")
    binary_target = cfg.get("binary_target") or cfg.get("binary_resp_nme")
    for col in [
        target,
        weight,
        binary_target,
        split_group_col,
        split_time_col,
        cv_group_col,
        cv_time_col,
        report_time_col,
    ]:
        if isinstance(col, str):
            cols.append(col)
    if report_group_cols:
        cols.extend([c for c in report_group_cols if isinstance(c, str)])
    return _dedupe_columns(cols)


def _load_and_split_dataset(
    *,
    deps: BayesOptRunnerDeps,
    data_path: Path,
    data_format: str,
    dtype_map: Dict[str, Any],
    required_columns: Optional[List[str]],
    use_stream_split: bool,
    holdout_ratio: float,
    rand_seed: Optional[int],
    stream_split_chunksize: int,
    split_strategy: str,
    split_group_col: Optional[str],
    split_time_col: Optional[str],
    split_time_ascending: bool,
) -> Tuple[pd.DataFrame, pd.DataFrame, int]:
    if use_stream_split:
        train_df, test_df, dataset_rows = _stream_random_split_csv(
            data_path,
            holdout_ratio=holdout_ratio,
            rand_seed=rand_seed,
            dtype_map=dtype_map,
            usecols=required_columns,
            low_memory=False,
            chunksize=int(stream_split_chunksize),
            coerce_dataset_types=deps.coerce_dataset_types,
        )
        return train_df, test_df, dataset_rows

    raw = deps.load_dataset(
        data_path,
        data_format=data_format,
        dtype_map=dtype_map,
        usecols=required_columns,
        low_memory=False,
    )
    train_df, test_df = deps.split_train_test(
        raw,
        holdout_ratio=holdout_ratio,
        strategy=split_strategy,
        group_col=split_group_col,
        time_col=split_time_col,
        time_ascending=split_time_ascending,
        rand_seed=rand_seed,
        reset_index_mode="time_group",
        ratio_label="holdout_ratio",
    )
    train_df = deps.coerce_dataset_types(train_df)
    test_df = deps.coerce_dataset_types(test_df)
    dataset_rows = int(len(raw))
    del raw
    return train_df, test_df, dataset_rows


def _resolve_ddp_preprocess_paths(
    *,
    output_dir: Optional[str],
    config_path: Path,
    model_name: str,
    config_sha: str,
) -> Tuple[Path, Path]:
    base_root = Path(output_dir) if output_dir else (config_path.parent / "Results")
    cache_root = base_root / "_ddp_preprocess_cache"
    cache_root.mkdir(parents=True, exist_ok=True)
    model_tag = "".join(ch if ch.isalnum() or ch in {"_", "-"} else "_" for ch in str(model_name))
    token_raw = f"{config_sha}|{model_name}"
    token = hashlib.sha256(token_raw.encode("utf-8")).hexdigest()[:16]
    meta_path = cache_root / f"{model_tag}_{token}.meta.json"
    bundle_path = cache_root / f"{model_tag}_{token}.bundle.pkl"
    return meta_path, bundle_path


def _write_json_atomic(path: Path, payload: Dict[str, Any]) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path = path.with_suffix(path.suffix + ".tmp")
    tmp_path.write_text(
        json.dumps(payload, ensure_ascii=True, sort_keys=True),
        encoding="utf-8",
    )
    tmp_path.replace(path)


def _read_json_with_retry(
    path: Path,
    *,
    retries: Optional[int] = None,
    delay_seconds: float = 0.1,
    timeout_seconds: Optional[float] = None,
) -> Dict[str, Any]:
    delay = max(0.01, float(delay_seconds))
    if timeout_seconds is None:
        raw_timeout = os.environ.get("BAYESOPT_DDP_META_READ_TIMEOUT_SECONDS", "300")
        try:
            timeout_seconds = float(raw_timeout)
        except Exception:
            timeout_seconds = 300.0
    timeout_seconds = max(delay, float(timeout_seconds))
    if retries is None:
        retries = max(1, int(timeout_seconds / delay))
    else:
        retries = max(1, int(retries))

    last_error: Optional[Exception] = None
    for attempt in range(retries):
        if path.exists():
            try:
                return json.loads(path.read_text(encoding="utf-8"))
            except Exception as exc:
                last_error = exc
        if attempt + 1 < retries:
            time.sleep(delay)
    waited_seconds = retries * delay
    if last_error is not None:
        raise RuntimeError(
            f"Failed to read JSON metadata after waiting {waited_seconds:.1f}s: "
            f"{path} ({last_error})"
        ) from last_error
    raise FileNotFoundError(
        f"Metadata file not found after waiting {waited_seconds:.1f}s: {path}"
    )


def run_bayesopt_entry_training(
    args: Any,
    *,
    script_dir: Path,
    deps: BayesOptRunnerDeps,
    hooks: BayesOptRunnerHooks,
    training_context_from_env: Callable[[], Any],
) -> None:
    config_path, cfg = deps.resolve_and_load_config(
        args.config_json,
        script_dir,
        required_keys=["data_dir", "model_list",
                       "model_categories", "target", "weight"],
    )
    plot_requested = bool(args.plot_curves or cfg.get("plot_curves", False))
    config_sha = hashlib.sha256(config_path.read_bytes()).hexdigest()
    run_id = datetime.utcnow().strftime("%Y%m%d_%H%M%S")

    dist_ctx = training_context_from_env()
    dist_rank = dist_ctx.rank
    dist_active = dist_ctx.is_distributed
    is_main_process = dist_ctx.is_main_process
    _ddp_barrier = _create_ddp_barrier(dist_ctx)

    data_dir, data_format, data_path_template, dtype_map = deps.resolve_data_config(
        cfg,
        config_path,
        create_data_dir=True,
    )
    runtime_cfg = deps.resolve_runtime_config(cfg)
    ddp_min_rows = runtime_cfg["ddp_min_rows"]
    bo_sample_limit = runtime_cfg["bo_sample_limit"]
    cache_predictions = runtime_cfg["cache_predictions"]
    prediction_cache_dir = runtime_cfg["prediction_cache_dir"]
    prediction_cache_format = runtime_cfg["prediction_cache_format"]
    report_cfg = deps.resolve_report_config(cfg)
    report_output_dir = report_cfg["report_output_dir"]
    report_group_cols = report_cfg["report_group_cols"]
    report_time_col = report_cfg["report_time_col"]
    report_time_freq = report_cfg["report_time_freq"]
    report_time_ascending = report_cfg["report_time_ascending"]
    psi_bins = report_cfg["psi_bins"]
    psi_strategy = report_cfg["psi_strategy"]
    psi_features = report_cfg["psi_features"]
    calibration_cfg = report_cfg["calibration_cfg"]
    threshold_cfg = report_cfg["threshold_cfg"]
    bootstrap_cfg = report_cfg["bootstrap_cfg"]
    register_model = report_cfg["register_model"]
    registry_path = report_cfg["registry_path"]
    registry_tags = report_cfg["registry_tags"]
    registry_status = report_cfg["registry_status"]
    data_fingerprint_max_bytes = report_cfg["data_fingerprint_max_bytes"]
    report_enabled = report_cfg["report_enabled"]

    split_cfg = deps.resolve_split_config(cfg)
    holdout_ratio = split_cfg["holdout_ratio"]
    val_ratio = split_cfg["val_ratio"]
    split_strategy = split_cfg["split_strategy"]
    split_group_col = split_cfg["split_group_col"]
    split_time_col = split_cfg["split_time_col"]
    split_time_ascending = split_cfg["split_time_ascending"]
    cv_strategy = split_cfg["cv_strategy"]
    cv_group_col = split_cfg["cv_group_col"]
    cv_time_col = split_cfg["cv_time_col"]
    cv_time_ascending = split_cfg["cv_time_ascending"]
    cv_splits = split_cfg["cv_splits"]
    ft_oof_folds = split_cfg["ft_oof_folds"]
    ft_oof_strategy = split_cfg["ft_oof_strategy"]
    ft_oof_shuffle = split_cfg["ft_oof_shuffle"]
    save_preprocess = runtime_cfg["save_preprocess"]
    preprocess_artifact_path = runtime_cfg["preprocess_artifact_path"]
    rand_seed = runtime_cfg["rand_seed"]
    epochs = runtime_cfg["epochs"]
    output_cfg = deps.resolve_output_dirs(
        cfg,
        config_path,
        output_override=args.output_dir,
    )
    output_dir = output_cfg["output_dir"]
    reuse_best_params = bool(
        args.reuse_best_params or runtime_cfg["reuse_best_params"])
    xgb_max_depth_max = runtime_cfg["xgb_max_depth_max"]
    xgb_n_estimators_max = runtime_cfg["xgb_n_estimators_max"]
    xgb_gpu_id = runtime_cfg["xgb_gpu_id"]
    xgb_cleanup_per_fold = runtime_cfg["xgb_cleanup_per_fold"]
    xgb_cleanup_synchronize = runtime_cfg["xgb_cleanup_synchronize"]
    xgb_use_dmatrix = runtime_cfg["xgb_use_dmatrix"]
    ft_cleanup_per_fold = runtime_cfg["ft_cleanup_per_fold"]
    ft_cleanup_synchronize = runtime_cfg["ft_cleanup_synchronize"]
    resn_cleanup_per_fold = runtime_cfg["resn_cleanup_per_fold"]
    resn_cleanup_synchronize = runtime_cfg["resn_cleanup_synchronize"]
    gnn_cleanup_per_fold = runtime_cfg["gnn_cleanup_per_fold"]
    gnn_cleanup_synchronize = runtime_cfg["gnn_cleanup_synchronize"]
    optuna_cleanup_synchronize = runtime_cfg["optuna_cleanup_synchronize"]
    optuna_storage = runtime_cfg["optuna_storage"]
    optuna_study_prefix = runtime_cfg["optuna_study_prefix"]
    best_params_files = runtime_cfg["best_params_files"]
    plot_path_style = runtime_cfg["plot_path_style"]
    stream_split_csv = runtime_cfg["stream_split_csv"]
    stream_split_chunksize = runtime_cfg["stream_split_chunksize"]

    model_names = deps.build_model_names(
        cfg["model_list"], cfg["model_categories"])
    if not model_names:
        raise ValueError(
            "No model names generated from model_list/model_categories.")

    results: Dict[str, Any] = {}
    trained_keys_by_model: Dict[str, List[str]] = {}

    for model_name in model_names:
        data_path = deps.resolve_data_path(
            data_dir,
            model_name,
            data_format=data_format,
            path_template=data_path_template,
        )
        if not data_path.exists():
            raise FileNotFoundError(f"Missing dataset: {data_path}")
        data_fingerprint = {"path": str(data_path)}
        if report_enabled and is_main_process:
            data_fingerprint = deps.fingerprint_file(
                data_path,
                max_bytes=data_fingerprint_max_bytes,
            )

        print(f"\n=== Processing model {model_name} ===")
        fmt_lower = str(data_format).strip().lower()
        is_csv_source = fmt_lower == "csv" or (
            fmt_lower == "auto" and data_path.suffix.lower() == ".csv"
        )
        resource_profile = str(
            cfg.get(
                "resource_profile",
                (cfg.get("env", {}) or {}).get("BAYESOPT_RESOURCE_PROFILE", "auto"),
            )
        ).strip().lower()
        auto_stream_split = bool(
            is_csv_source
            and str(split_strategy).strip().lower() in {"random"}
            and resource_profile == "memory_saving"
        )
        use_stream_split = bool(
            (stream_split_csv or auto_stream_split)
            and is_csv_source
            and str(split_strategy).strip().lower() in {"random"}
        )
        if auto_stream_split and not stream_split_csv:
            print(
                "[Data] Auto-enabling stream_split_csv for memory_saving profile.",
                flush=True,
            )
        required_columns = _resolve_required_columns(
            cfg,
            split_group_col=split_group_col,
            split_time_col=split_time_col,
            cv_group_col=cv_group_col,
            cv_time_col=cv_time_col,
            report_group_cols=report_group_cols,
            report_time_col=report_time_col,
        )
        if required_columns:
            print(
                f"[Data] projected column loading enabled: {len(required_columns)} columns",
                flush=True,
            )

        ddp_requested_by_config = bool(
            args.use_resn_ddp
            or cfg.get("use_resn_ddp", False)
            or args.use_ft_ddp
            or cfg.get("use_ft_ddp", False)
        )
        skip_data_load_for_non_main = bool(
            dist_active
            and dist_rank != 0
            and (
                (not ddp_requested_by_config)
                or (not bool(cfg.get("use_gpu", True)))
            )
        )
        use_shared_preprocess = bool(
            dist_active
            and cfg.get("use_gpu", True)
            and ddp_requested_by_config
        )
        shared_meta_path: Optional[Path] = None
        shared_bundle_path: Optional[Path] = None
        train_df: Optional[pd.DataFrame] = None
        test_df: Optional[pd.DataFrame] = None
        dataset_rows = 0
        ddp_enabled: Optional[bool] = None

        if skip_data_load_for_non_main:
            ddp_enabled = False
            print(
                f"[Data][Rank {dist_rank}] Skip dataset load because no DDP trainer is requested.",
                flush=True,
            )

        if use_shared_preprocess:
            shared_meta_path, shared_bundle_path = _resolve_ddp_preprocess_paths(
                output_dir=output_dir,
                config_path=config_path,
                model_name=model_name,
                config_sha=config_sha,
            )
            if dist_rank == 0:
                if use_stream_split:
                    print(
                        f"[Data] streaming random split enabled "
                        f"(chunksize={int(stream_split_chunksize)}) for {data_path}",
                        flush=True,
                    )
                train_df, test_df, dataset_rows = _load_and_split_dataset(
                    deps=deps,
                    data_path=data_path,
                    data_format=data_format,
                    dtype_map=dtype_map,
                    required_columns=required_columns,
                    use_stream_split=use_stream_split,
                    holdout_ratio=holdout_ratio,
                    rand_seed=rand_seed,
                    stream_split_chunksize=int(stream_split_chunksize),
                    split_strategy=split_strategy,
                    split_group_col=split_group_col,
                    split_time_col=split_time_col,
                    split_time_ascending=split_time_ascending,
                )
                ddp_enabled = bool(
                    dist_active
                    and cfg.get("use_gpu", True)
                    and (dataset_rows >= int(ddp_min_rows))
                )
                _write_json_atomic(
                    shared_meta_path,
                    {
                        "dataset_rows": int(dataset_rows),
                        "ddp_enabled": bool(ddp_enabled),
                        "bundle_path": str(shared_bundle_path),
                    },
                )
            _ddp_barrier(f"shared_preprocess_meta_ready_{model_name}")
            if dist_rank != 0:
                meta = _read_json_with_retry(shared_meta_path)
                dataset_rows = int(meta.get("dataset_rows", 0))
                ddp_enabled = bool(meta.get("ddp_enabled", False))
                if ddp_enabled:
                    print(
                        f"[Data][Rank {dist_rank}] Skip raw split/load; "
                        f"will load preprocess bundle: {shared_bundle_path}",
                        flush=True,
                    )
                else:
                    skip_data_load_for_non_main = True
                    print(
                        f"[Data][Rank {dist_rank}] Skip dataset load because DDP is disabled "
                        f"(dataset_rows={dataset_rows}, ddp_min_rows={int(ddp_min_rows)}).",
                        flush=True,
                    )
        else:
            if not skip_data_load_for_non_main:
                if use_stream_split:
                    print(
                        f"[Data] streaming random split enabled "
                        f"(chunksize={int(stream_split_chunksize)}) for {data_path}",
                        flush=True,
                    )
                train_df, test_df, dataset_rows = _load_and_split_dataset(
                    deps=deps,
                    data_path=data_path,
                    data_format=data_format,
                    dtype_map=dtype_map,
                    required_columns=required_columns,
                    use_stream_split=use_stream_split,
                    holdout_ratio=holdout_ratio,
                    rand_seed=rand_seed,
                    stream_split_chunksize=int(stream_split_chunksize),
                    split_strategy=split_strategy,
                    split_group_col=split_group_col,
                    split_time_col=split_time_col,
                    split_time_ascending=split_time_ascending,
                )

        use_resn_dp = bool((args.use_resn_dp or cfg.get(
            "use_resn_data_parallel", False)) and cfg.get("use_gpu", True))
        use_ft_dp = bool((args.use_ft_dp or cfg.get(
            "use_ft_data_parallel", False)) and cfg.get("use_gpu", True))
        if ddp_enabled is None:
            ddp_enabled = bool(
                dist_active
                and cfg.get("use_gpu", True)
                and (dataset_rows >= int(ddp_min_rows))
            )
        use_resn_ddp = (args.use_resn_ddp or cfg.get(
            "use_resn_ddp", False)) and ddp_enabled
        use_ft_ddp = (args.use_ft_ddp or cfg.get(
            "use_ft_ddp", False)) and ddp_enabled
        use_gnn_dp = bool((args.use_gnn_dp or cfg.get(
            "use_gnn_data_parallel", False)) and cfg.get("use_gpu", True))
        gnn_use_ann = cfg.get("gnn_use_approx_knn", True)
        if args.gnn_no_ann:
            gnn_use_ann = False
        gnn_threshold = args.gnn_ann_threshold if args.gnn_ann_threshold is not None else cfg.get(
            "gnn_approx_knn_threshold", 50000)
        gnn_graph_cache = args.gnn_graph_cache or cfg.get("gnn_graph_cache")
        if isinstance(gnn_graph_cache, str) and gnn_graph_cache.strip():
            resolved_cache = deps.resolve_path(gnn_graph_cache, config_path.parent)
            if resolved_cache is not None:
                gnn_graph_cache = str(resolved_cache)
        gnn_max_gpu_nodes = args.gnn_max_gpu_nodes if args.gnn_max_gpu_nodes is not None else cfg.get(
            "gnn_max_gpu_knn_nodes", 200000)
        gnn_gpu_mem_ratio = args.gnn_gpu_mem_ratio if args.gnn_gpu_mem_ratio is not None else cfg.get(
            "gnn_knn_gpu_mem_ratio", 0.9)
        gnn_gpu_mem_overhead = args.gnn_gpu_mem_overhead if args.gnn_gpu_mem_overhead is not None else cfg.get(
            "gnn_knn_gpu_mem_overhead", 2.0)

        binary_target = cfg.get("binary_target") or cfg.get("binary_resp_nme")
        task_type = str(cfg.get("task_type", "regression"))
        feature_list = cfg.get("feature_list")
        categorical_features = cfg.get("categorical_features")
        use_gpu = bool(cfg.get("use_gpu", True))
        region_province_col = cfg.get("region_province_col")
        region_city_col = cfg.get("region_city_col")
        region_effect_alpha = cfg.get("region_effect_alpha")
        geo_feature_nmes = cfg.get("geo_feature_nmes")
        geo_token_hidden_dim = cfg.get("geo_token_hidden_dim")
        geo_token_layers = cfg.get("geo_token_layers")
        geo_token_dropout = cfg.get("geo_token_dropout")
        geo_token_k_neighbors = cfg.get("geo_token_k_neighbors")
        geo_token_learning_rate = cfg.get("geo_token_learning_rate")
        geo_token_epochs = cfg.get("geo_token_epochs")

        ft_role = args.ft_role or cfg.get("ft_role", "model")
        if args.ft_as_feature and args.ft_role is None:
            if str(cfg.get("ft_role", "model")) == "model":
                ft_role = "embedding"
        ft_feature_prefix = str(
            cfg.get("ft_feature_prefix", args.ft_feature_prefix))
        ft_num_numeric_tokens = cfg.get("ft_num_numeric_tokens")

        config_fields = getattr(
            deps.ropt.BayesOptConfig,
            "__dataclass_fields__",
            {},
        )
        allowed_config_keys = {
            key
            for key, spec in config_fields.items()
            if getattr(spec, "init", True)
        }
        config_payload = {
            k: v for k, v in cfg.items() if k in allowed_config_keys and v is not None
        }
        config_payload.update({
            k: v
            for k, v in runtime_cfg.items()
            if k in allowed_config_keys and v is not None
        })
        config_payload.update({
            k: v
            for k, v in split_cfg.items()
            if k in allowed_config_keys and v is not None
        })
        save_preprocess_bundle = bool(
            use_shared_preprocess
            and ddp_enabled
            and dist_rank == 0
            and shared_bundle_path is not None
        )
        load_preprocess_bundle = bool(
            use_shared_preprocess
            and ddp_enabled
            and dist_rank != 0
            and shared_bundle_path is not None
        )
        override_payload = {
            "model_nme": model_name,
            "resp_nme": cfg["target"],
            "weight_nme": cfg["weight"],
            "factor_nmes": feature_list,
            "task_type": task_type,
            "binary_resp_nme": binary_target,
            "cate_list": categorical_features,
            "prop_test": val_ratio,
            "rand_seed": rand_seed,
            "epochs": epochs,
            "use_gpu": use_gpu,
            "use_resn_data_parallel": use_resn_dp,
            "use_ft_data_parallel": use_ft_dp,
            "use_gnn_data_parallel": use_gnn_dp,
            "use_resn_ddp": use_resn_ddp,
            "use_ft_ddp": use_ft_ddp,
            "output_dir": output_dir,
            "xgb_max_depth_max": xgb_max_depth_max,
            "xgb_n_estimators_max": xgb_n_estimators_max,
            "xgb_gpu_id": xgb_gpu_id,
            "xgb_cleanup_per_fold": xgb_cleanup_per_fold,
            "xgb_cleanup_synchronize": xgb_cleanup_synchronize,
            "xgb_use_dmatrix": xgb_use_dmatrix,
            "ft_cleanup_per_fold": ft_cleanup_per_fold,
            "ft_cleanup_synchronize": ft_cleanup_synchronize,
            "resn_cleanup_per_fold": resn_cleanup_per_fold,
            "resn_cleanup_synchronize": resn_cleanup_synchronize,
            "gnn_cleanup_per_fold": gnn_cleanup_per_fold,
            "gnn_cleanup_synchronize": gnn_cleanup_synchronize,
            "optuna_cleanup_synchronize": optuna_cleanup_synchronize,
            "resn_weight_decay": cfg.get("resn_weight_decay"),
            "final_ensemble": bool(cfg.get("final_ensemble", False)),
            "final_ensemble_k": int(cfg.get("final_ensemble_k", 3)),
            "final_refit": bool(cfg.get("final_refit", True)),
            "optuna_storage": optuna_storage,
            "optuna_study_prefix": optuna_study_prefix,
            "best_params_files": best_params_files,
            "gnn_use_approx_knn": gnn_use_ann,
            "gnn_approx_knn_threshold": gnn_threshold,
            "gnn_graph_cache": gnn_graph_cache,
            "gnn_max_gpu_knn_nodes": gnn_max_gpu_nodes,
            "gnn_knn_gpu_mem_ratio": gnn_gpu_mem_ratio,
            "gnn_knn_gpu_mem_overhead": gnn_gpu_mem_overhead,
            "region_province_col": region_province_col,
            "region_city_col": region_city_col,
            "region_effect_alpha": region_effect_alpha,
            "geo_feature_nmes": geo_feature_nmes,
            "geo_token_hidden_dim": geo_token_hidden_dim,
            "geo_token_layers": geo_token_layers,
            "geo_token_dropout": geo_token_dropout,
            "geo_token_k_neighbors": geo_token_k_neighbors,
            "geo_token_learning_rate": geo_token_learning_rate,
            "geo_token_epochs": geo_token_epochs,
            "ft_role": ft_role,
            "ft_feature_prefix": ft_feature_prefix,
            "ft_num_numeric_tokens": ft_num_numeric_tokens,
            "reuse_best_params": reuse_best_params,
            "bo_sample_limit": bo_sample_limit,
            "cache_predictions": cache_predictions,
            "prediction_cache_dir": prediction_cache_dir,
            "prediction_cache_format": prediction_cache_format,
            "cv_strategy": cv_strategy or split_strategy,
            "cv_group_col": cv_group_col or split_group_col,
            "cv_time_col": cv_time_col or split_time_col,
            "cv_time_ascending": cv_time_ascending,
            "cv_splits": cv_splits,
            "ft_oof_folds": ft_oof_folds,
            "ft_oof_strategy": ft_oof_strategy,
            "ft_oof_shuffle": ft_oof_shuffle,
            "save_preprocess": save_preprocess,
            "preprocess_artifact_path": preprocess_artifact_path,
            "save_preprocess_bundle": save_preprocess_bundle,
            "load_preprocess_bundle": load_preprocess_bundle,
            "preprocess_bundle_path": (
                str(shared_bundle_path)
                if use_shared_preprocess and ddp_enabled and shared_bundle_path is not None
                else None
            ),
            "plot_path_style": plot_path_style or "nested",
        }
        config_payload.update({
            k: v
            for k, v in override_payload.items()
            if k in allowed_config_keys and v is not None
        })
        config = deps.ropt.BayesOptConfig.from_flat_dict(config_payload)

        if "all" in args.model_keys:
            requested_keys = ["glm", "xgb", "resn", "ft", "gnn"]
        else:
            requested_keys = args.model_keys
        requested_keys = deps.dedupe_preserve_order(requested_keys)

        if ft_role != "model":
            requested_keys = [k for k in requested_keys if k != "ft"]
            if not requested_keys:
                stack_keys = args.stack_model_keys or cfg.get(
                    "stack_model_keys")
                if stack_keys:
                    if "all" in stack_keys:
                        requested_keys = ["glm", "xgb", "resn", "gnn"]
                    else:
                        requested_keys = [k for k in stack_keys if k != "ft"]
                    requested_keys = deps.dedupe_preserve_order(requested_keys)
        known_model_keys = {"glm", "xgb", "resn", "ft", "gnn"}
        invalid_requested = [k for k in requested_keys if k not in known_model_keys]
        if invalid_requested:
            raise ValueError(
                f"Unknown requested model key(s): {invalid_requested}. "
                f"Valid keys: {sorted(known_model_keys)}"
            )

        # In torchrun non-DDP mode, non-main ranks only coordinate barriers and
        # skip all heavy dataset/model construction.
        if dist_active and not ddp_enabled and dist_rank != 0:
            if ft_role != "model":
                _ddp_barrier("start_ft_embedding")
                _ddp_barrier("finish_ft_embedding")
            for key in requested_keys:
                _ddp_barrier(f"start_non_ddp_{model_name}_{key}")
                _ddp_barrier(f"finish_non_ddp_{model_name}_{key}")
            continue

        if use_shared_preprocess and ddp_enabled and shared_bundle_path is not None and dist_active:
            if dist_rank == 0:
                model = deps.ropt.BayesOptModel(train_df, test_df, config=config)
                _ddp_barrier(f"shared_preprocess_bundle_ready_{model_name}")
            else:
                _ddp_barrier(f"shared_preprocess_bundle_ready_{model_name}")
                model = deps.ropt.BayesOptModel(None, None, config=config)
        else:
            model = deps.ropt.BayesOptModel(train_df, test_df, config=config)

        if plot_requested:
            plot_cfg = cfg.get("plot", {})
            plot_enabled = bool(plot_cfg.get("enable", False))
            if plot_enabled and plot_cfg.get("pre_oneway", False) and plot_cfg.get("oneway", True):
                n_bins = int(plot_cfg.get("n_bins", 10))
                model.plot_oneway(n_bins=n_bins, plot_subdir="oneway/pre")

        if ft_role != "model" and dist_active and ddp_enabled:
            ft_trainer = model.trainers.get("ft")
            if ft_trainer is None:
                raise ValueError("FT trainer is not available.")
            ft_trainer_uses_ddp = bool(
                getattr(ft_trainer, "enable_distributed_optuna", False))
            if not ft_trainer_uses_ddp:
                raise ValueError(
                    "FT embedding under torchrun requires enabling FT DDP (use --use-ft-ddp or set use_ft_ddp=true)."
                )
        missing = [key for key in requested_keys if key not in model.trainers]
        if missing:
            raise ValueError(
                f"Trainer(s) {missing} not available for {model_name}")

        executed_keys: List[str] = []
        if ft_role != "model":
            if dist_active and not ddp_enabled:
                _ddp_barrier("start_ft_embedding")
                if dist_rank != 0:
                    _ddp_barrier("finish_ft_embedding")
                    continue
            print(
                f"Optimizing ft as {ft_role} for {model_name} (max_evals={args.max_evals})")
            model.optimize_model("ft", max_evals=args.max_evals)
            model.trainers["ft"].save()
            _free_cuda_safe(deps.ropt)
            if dist_active and not ddp_enabled:
                _ddp_barrier("finish_ft_embedding")
        for key in requested_keys:
            trainer = model.trainers[key]
            trainer_uses_ddp = bool(
                getattr(trainer, "enable_distributed_optuna", False))
            if dist_active and not trainer_uses_ddp:
                if dist_rank != 0:
                    print(
                        f"[Rank {dist_rank}] Skip {model_name}/{key} because trainer is not DDP-enabled."
                    )
                _ddp_barrier(f"start_non_ddp_{model_name}_{key}")
                if dist_rank != 0:
                    _ddp_barrier(f"finish_non_ddp_{model_name}_{key}")
                    continue

            print(
                f"Optimizing {key} for {model_name} (max_evals={args.max_evals})")
            model.optimize_model(key, max_evals=args.max_evals)
            model.trainers[key].save()
            hooks.plot_loss_curve_for_trainer(model_name, model.trainers[key])
            if key in deps.pytorch_trainers:
                _free_cuda_safe(deps.ropt)
            if dist_active and not trainer_uses_ddp:
                _ddp_barrier(f"finish_non_ddp_{model_name}_{key}")
            executed_keys.append(key)

        if not executed_keys:
            continue

        results[model_name] = model
        trained_keys_by_model[model_name] = executed_keys
        if report_enabled and is_main_process:
            psi_report_df = hooks.compute_psi_report(
                model,
                features=psi_features,
                bins=psi_bins,
                strategy=str(psi_strategy),
            )
            for key in executed_keys:
                hooks.evaluate_and_report(
                    model,
                    model_name=model_name,
                    model_key=key,
                    cfg=cfg,
                    data_path=data_path,
                    data_fingerprint=data_fingerprint,
                    report_output_dir=report_output_dir,
                    report_group_cols=report_group_cols,
                    report_time_col=report_time_col,
                    report_time_freq=str(report_time_freq),
                    report_time_ascending=bool(report_time_ascending),
                    psi_report_df=psi_report_df,
                    calibration_cfg=calibration_cfg,
                    threshold_cfg=threshold_cfg,
                    bootstrap_cfg=bootstrap_cfg,
                    register_model=register_model,
                    registry_path=registry_path,
                    registry_tags=registry_tags,
                    registry_status=registry_status,
                    run_id=run_id,
                    config_sha=config_sha,
                )

    if not plot_requested:
        return

    for name, model in results.items():
        hooks.plot_curves_for_model(
            model,
            trained_keys_by_model.get(name, []),
            cfg,
        )


__all__ = [
    "BayesOptRunnerDeps",
    "BayesOptRunnerHooks",
    "run_bayesopt_entry_training",
]



"""Tests for pricing factors module."""

import numpy as np
import pandas as pd
import pytest

from ins_pricing.exceptions import DataValidationError


@pytest.fixture
def sample_policy_data():
    """Sample insurance policy data."""
    np.random.seed(42)
    return pd.DataFrame({
        "policy_id": range(1000),
        "age": np.random.randint(18, 80, 1000),
        "gender": np.random.choice(["M", "F"], 1000),
        "region": np.random.choice(["North", "South", "East", "West"], 1000),
        "vehicle_age": np.random.randint(0, 15, 1000),
        "claim_amount": np.random.exponential(500, 1000),
        "exposure": np.random.uniform(0.5, 1.0, 1000),
        "premium": np.random.uniform(200, 1000, 1000)
    })


class TestFactorTableConstruction:
    """Test factor table construction."""

    def test_build_univariate_factor_table(self, sample_policy_data):
        """Test building a univariate factor table."""
        from ins_pricing.pricing.factors import build_factor_table

        factor_table = build_factor_table(
            df=sample_policy_data,
            factor_col="age",
            loss_col="claim_amount",
            exposure_col="exposure",
            method="quantile",
            n_bins=10
        )

        assert len(factor_table) <= 10
        assert "age_bin" in factor_table.columns
        assert "relativity" in factor_table.columns
        assert "claim_count" in factor_table.columns

    def test_equal_width_binning(self, sample_policy_data):
        """Test equal width binning strategy."""
        from ins_pricing.pricing.factors import build_factor_table

        factor_table = build_factor_table(
            df=sample_policy_data,
            factor_col="vehicle_age",
            loss_col="claim_amount",
            exposure_col="exposure",
            method="equal_width",
            n_bins=5
        )

        assert len(factor_table) == 5

    def test_categorical_factor_table(self, sample_policy_data):
        """Test factor table for categorical variables."""
        from ins_pricing.pricing.factors import build_factor_table

        factor_table = build_factor_table(
            df=sample_policy_data,
            factor_col="region",
            loss_col="claim_amount",
            exposure_col="exposure",
            method="categorical"
        )

        assert set(factor_table["region"]) == set(sample_policy_data["region"].unique())
        assert "relativity" in factor_table.columns


class TestFactorSmoothing:
    """Test factor smoothing techniques."""

    def test_credibility_weighting(self):
        """Test credibility-weighted smoothing."""
        from ins_pricing.pricing.factors import apply_credibility_smoothing

        raw_factors = pd.DataFrame({
            "bin": ["A", "B", "C"],
            "relativity": [1.2, 0.8, 1.5],
            "exposure": [100, 500, 50]  # C has low credibility
        })

        smoothed = apply_credibility_smoothing(raw_factors, base_relativity=1.0)

        # Low exposure bin should be pulled toward base
        assert abs(smoothed.loc[2, "relativity"] - 1.0) < abs(raw_factors.loc[2, "relativity"] - 1.0)

    def test_neighbor_smoothing(self):
        """Test smoothing using neighboring bins."""
        from ins_pricing.pricing.factors import apply_neighbor_smoothing

        factors = pd.DataFrame({
            "bin": [1, 2, 3, 4, 5],
            "relativity": [1.0, 1.2, 2.5, 1.4, 1.5]  # Bin 3 is outlier
        })

        smoothed = apply_neighbor_smoothing(factors)

        # Outlier should be smoothed
        assert smoothed.loc[2, "relativity"] < factors.loc[2, "relativity"]


class TestFactorApplication:
    """Test applying factors to new data."""

    def test_apply_factors_to_policies(self, sample_policy_data):
        """Test applying factor table to policies."""
        from ins_pricing.pricing.factors import build_factor_table, apply_factors

        # Build factor table
        age_factors = build_factor_table(
            df=sample_policy_data,
            factor_col="age",
            loss_col="claim_amount",
            exposure_col="exposure",
            n_bins=5
        )

        # Apply to new data
        result = apply_factors(sample_policy_data, age_factors, factor_col="age")

        assert "age_relativity" in result.columns
        assert result["age_relativity"].notna().all()


@pytest.mark.parametrize("method,n_bins", [
    ("quantile", 5),
    ("quantile", 10),
    ("equal_width", 5),
    ("equal_width", 10),
])
class TestBinningMethods:
    """Test different binning methods."""

    def test_binning_produces_expected_bins(self, sample_policy_data, method, n_bins):
        """Test that binning produces expected number of bins."""
        from ins_pricing.pricing.factors import build_factor_table

        factor_table = build_factor_table(
            df=sample_policy_data,
            factor_col="age",
            loss_col="claim_amount",
            exposure_col="exposure",
            method=method,
            n_bins=n_bins
        )

        assert len(factor_table) <= n_bins



from __future__ import annotations

import numpy as np
import pandas as pd

from ins_pricing.utils.paths import coerce_dataset_types, load_dataset


def test_load_dataset_projects_columns_and_dtype_map(tmp_path):
    path = tmp_path / "sample.csv"
    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": [0.1, 0.2, 0.3],
            "c": ["x", "y", "z"],
        }
    )
    df.to_csv(path, index=False)

    out = load_dataset(
        path,
        data_format="csv",
        dtype_map={"a": "int32", "b": "float32", "missing": "float32"},
        usecols=["a", "c"],
        low_memory=False,
    )

    assert list(out.columns) == ["a", "c"]
    assert str(out["a"].dtype) in {"int32", "int64"}


def test_coerce_dataset_types_is_in_place_and_downcasts_numeric():
    df = pd.DataFrame(
        {
            "num_i": np.array([1, 2, 3], dtype=np.int64),
            "num_f": np.array([1.0, np.nan, 3.5], dtype=np.float64),
            "cat": ["a", None, "c"],
        }
    )

    out = coerce_dataset_types(df)

    assert out is df
    assert out["num_i"].dtype == np.float32
    assert out["num_f"].dtype == np.float32
    assert out["num_f"].isna().sum() == 0
    assert out["cat"].tolist() == ["a", "<NA>", "c"]


def test_load_dataset_with_chunksize_returns_iterator(tmp_path):
    path = tmp_path / "sample_chunks.csv"
    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4],
            "b": [0.1, 0.2, 0.3, 0.4],
        }
    )
    df.to_csv(path, index=False)

    chunks = load_dataset(
        path,
        data_format="csv",
        usecols=["a", "b"],
        low_memory=False,
        chunksize=2,
    )

    assert not isinstance(chunks, pd.DataFrame)
    first = next(iter(chunks))
    assert list(first.columns) == ["a", "b"]
    assert len(first) == 2



"""Tests for production inference module (lightweight API checks)."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("ins_pricing.production.inference", reason="inference module not available")

from ins_pricing.production.inference import (
    ModelSpec,
    Predictor,
    PredictorRegistry,
    load_predictor,
    register_model_loader,
)


@dataclass
class _DummyPredictor(Predictor):
    value: float = 1.0

    def predict(self, df):  # type: ignore[override]
        return np.full(len(df), self.value)


def test_registry_loads_custom_predictor():
    registry = PredictorRegistry()
    captured = {}

    def _loader(spec: ModelSpec) -> Predictor:
        captured["spec"] = spec
        return _DummyPredictor(value=3.0)

    register_model_loader("xgb", _loader, registry=registry)
    spec = ModelSpec(
        model_key="xgb",
        model_name="demo",
        task_type="regression",
        cfg={},
        output_dir=Path("."),
        artifacts=None,
    )
    predictor = load_predictor(spec, registry=registry)
    assert isinstance(predictor, _DummyPredictor)
    assert captured["spec"] is spec


def test_registry_missing_key_raises():
    registry = PredictorRegistry()
    spec = ModelSpec(
        model_key="glm",
        model_name="demo",
        task_type="regression",
        cfg={},
        output_dir=Path("."),
        artifacts=None,
    )
    with pytest.raises(KeyError):
        load_predictor(spec, registry=registry)


def test_register_overwrite_controls():
    registry = PredictorRegistry()

    def _loader_one(spec: ModelSpec) -> Predictor:
        return _DummyPredictor(value=1.0)

    def _loader_two(spec: ModelSpec) -> Predictor:
        return _DummyPredictor(value=2.0)

    register_model_loader("ft", _loader_one, registry=registry)
    with pytest.raises(ValueError):
        register_model_loader("ft", _loader_two, registry=registry, overwrite=False)

    register_model_loader("ft", _loader_two, registry=registry, overwrite=True)
    spec = ModelSpec(
        model_key="ft",
        model_name="demo",
        task_type="regression",
        cfg={},
        output_dir=Path("."),
        artifacts=None,
    )
    predictor = load_predictor(spec, registry=registry)
    assert isinstance(predictor, _DummyPredictor)
    assert predictor.value == 2.0



"""Plotting workflows used by the frontend UI."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Optional, Tuple

import pandas as pd

from ins_pricing.cli.utils.cli_common import split_train_test
from ins_pricing.modelling.plotting import (
    PlotStyle,
    plot_double_lift_curve,
    plot_lift_curve,
    plot_oneway,
)
from ins_pricing.modelling.plotting.common import finalize_figure, plt
from ins_pricing.production.inference import load_predictor_from_config

from .workflows_common import (
    _dedupe_list,
    _drop_duplicate_columns,
    _infer_categorical_features,
    _parse_csv_list,
    _resolve_output_dir,
    _resolve_plot_path,
    _resolve_plot_style,
    _safe_tag,
)


def run_pre_oneway(
    *,
    data_path: str,
    model_name: str,
    target_col: str,
    weight_col: str,
    feature_list: str,
    categorical_features: str,
    n_bins: int = 10,
    holdout_ratio: Optional[float] = 0.25,
    rand_seed: int = 13,
    output_dir: Optional[str] = None,
) -> str:
    data_path = str(data_path or "").strip()
    if not data_path:
        raise ValueError("data_path is required.")
    model_name = str(model_name or "").strip()
    if not model_name:
        raise ValueError("model_name is required.")

    raw_path = Path(data_path).resolve()
    if not raw_path.exists():
        raise FileNotFoundError(f"Data file not found: {raw_path}")

    raw = pd.read_csv(raw_path, low_memory=False)
    raw = _drop_duplicate_columns(raw, "raw").reset_index(drop=True)
    raw.fillna(0, inplace=True)

    features = _dedupe_list(_parse_csv_list(feature_list))
    cats = _dedupe_list(_parse_csv_list(categorical_features))

    if not features:
        raise ValueError("feature_list is empty.")

    missing = [f for f in features if f not in raw.columns]
    if missing:
        print(f"[Warn] Missing features removed: {missing}")
        features = [f for f in features if f in raw.columns]
        cats = [f for f in cats if f in raw.columns]

    if not cats:
        cats = _infer_categorical_features(raw, features)

    out_dir = (
        Path(output_dir).resolve()
        if output_dir
        else raw_path.parent / "Results" / "plot" / model_name / "oneway" / "pre"
    )
    out_dir.mkdir(parents=True, exist_ok=True)

    if holdout_ratio is not None and float(holdout_ratio) > 0:
        train_df, _ = split_train_test(
            raw,
            holdout_ratio=float(holdout_ratio),
            strategy="random",
            rand_seed=int(rand_seed),
            reset_index_mode="none",
            ratio_label="holdout_ratio",
        )
        df = train_df.reset_index(drop=True).copy()
    else:
        df = raw.copy()

    print(f"Generating oneway plots for {len(features)} features...")
    saved = 0
    for i, feature in enumerate(features, 1):
        is_categorical = feature in cats
        try:
            save_path = out_dir / f"{feature}.png"
            plot_oneway(
                df,
                feature=feature,
                weight_col=weight_col,
                target_col=target_col,
                n_bins=int(n_bins),
                is_categorical=is_categorical,
                save_path=str(save_path),
                show=False,
            )
            if save_path.exists():
                saved += 1
            if i % 5 == 0 or i == len(features):
                print(f"  [{i}/{len(features)}] {feature}")
        except Exception as exc:
            print(f"  [Warn] {feature} failed: {exc}")

    print(f"Complete. Saved {saved}/{len(features)} plots to: {out_dir}")
    return f"Saved {saved} plots to {out_dir}"


def run_plot_direct(
    *,
    cfg_path: str,
    xgb_cfg_path: str,
    resn_cfg_path: str,
) -> str:
    cfg_path = Path(cfg_path).resolve()
    xgb_cfg_path = Path(xgb_cfg_path).resolve()
    resn_cfg_path = Path(resn_cfg_path).resolve()

    cfg = json.loads(cfg_path.read_text(encoding="utf-8"))
    xgb_cfg = json.loads(xgb_cfg_path.read_text(encoding="utf-8"))
    resn_cfg = json.loads(resn_cfg_path.read_text(encoding="utf-8"))

    model_name = f"{cfg['model_list'][0]}_{cfg['model_categories'][0]}"

    raw_data_dir = (cfg_path.parent / cfg["data_dir"]).resolve()
    raw_path = raw_data_dir / f"{model_name}.csv"
    raw = pd.read_csv(raw_path)
    raw = _drop_duplicate_columns(raw, "raw").reset_index(drop=True)
    raw.fillna(0, inplace=True)

    holdout_ratio = cfg.get("holdout_ratio", cfg.get("prop_test", 0.25))
    split_strategy = cfg.get("split_strategy", "random")
    split_group_col = cfg.get("split_group_col")
    split_time_col = cfg.get("split_time_col")
    split_time_ascending = cfg.get("split_time_ascending", True)
    rand_seed = cfg.get("rand_seed", 13)

    train_raw, test_raw = split_train_test(
        raw,
        holdout_ratio=holdout_ratio,
        strategy=split_strategy,
        group_col=split_group_col,
        time_col=split_time_col,
        time_ascending=split_time_ascending,
        rand_seed=rand_seed,
        reset_index_mode="none",
        ratio_label="holdout_ratio",
    )
    train_raw = _drop_duplicate_columns(train_raw, "train_raw")
    test_raw = _drop_duplicate_columns(test_raw, "test_raw")

    train_df = train_raw.copy()
    test_df = test_raw.copy()

    feature_list = _dedupe_list(cfg.get("feature_list") or [])
    categorical_features = _dedupe_list(cfg.get("categorical_features") or [])

    output_dir_map = {
        "xgb": _resolve_output_dir(xgb_cfg, xgb_cfg_path),
        "resn": _resolve_output_dir(resn_cfg, resn_cfg_path),
    }
    plot_path_style_map = {
        "xgb": _resolve_plot_style(xgb_cfg),
        "resn": _resolve_plot_style(resn_cfg),
    }

    def _get_plot_config(model_key: str) -> Tuple[str, str]:
        return (
            output_dir_map.get(model_key, _resolve_output_dir(cfg, cfg_path)),
            plot_path_style_map.get(model_key, _resolve_plot_style(cfg)),
        )

    def _load_predictor(cfg_path: Path, model_key: str):
        return load_predictor_from_config(cfg_path, model_key, model_name=model_name)

    model_cfg_map = {"xgb": xgb_cfg_path, "resn": resn_cfg_path}
    model_keys = cfg.get("model_keys") or ["xgb", "resn"]
    model_keys = [key for key in model_keys if key in model_cfg_map]
    if not model_keys:
        raise ValueError("No valid model keys found in plot config.")

    default_model_labels = {"xgb": "Xgboost", "resn": "ResNet"}

    def _model_label(model_key: str) -> str:
        labels = cfg.get("model_labels") or {}
        return str(labels.get(model_key, default_model_labels.get(model_key, model_key)))

    predictors = {key: _load_predictor(model_cfg_map[key], key) for key in model_keys}

    pred_train = {}
    pred_test = {}
    for key, predictor in predictors.items():
        pred_train[key] = predictor.predict(train_df).reshape(-1)
        pred_test[key] = predictor.predict(test_df).reshape(-1)
        if len(pred_train[key]) != len(train_df):
            raise ValueError(f"Train prediction length mismatch for {key}")
        if len(pred_test[key]) != len(test_df):
            raise ValueError(f"Test prediction length mismatch for {key}")

    plot_train = train_raw.copy()
    plot_test = test_raw.copy()
    for key in model_keys:
        plot_train[f"pred_{key}"] = pred_train[key]
        plot_test[f"pred_{key}"] = pred_test[key]

    weight_col = cfg["weight"]
    target_col = cfg["target"]

    if weight_col not in plot_train.columns:
        plot_train[weight_col] = 1.0
    if weight_col not in plot_test.columns:
        plot_test[weight_col] = 1.0
    if target_col in plot_train.columns:
        plot_train["w_act"] = plot_train[target_col] * plot_train[weight_col]
    if target_col in plot_test.columns:
        plot_test["w_act"] = plot_test[target_col] * plot_test[weight_col]

    if "w_act" not in plot_train.columns or plot_train["w_act"].isna().all():
        print("[Plot] Missing target values in train split; skip plots.")
        return "Skipped plotting due to missing target values."

    n_bins = cfg.get("plot", {}).get("n_bins", 10)
    oneway_features = feature_list
    oneway_categorical = set(categorical_features)

    for pred_key in model_keys:
        pred_label = _model_label(pred_key)
        pred_col = f"pred_{pred_key}"
        pred_tag = _safe_tag(pred_label or pred_col)
        output_root, plot_style = _get_plot_config(pred_key)
        for feature in oneway_features:
            if feature not in plot_train.columns:
                continue
            save_path = _resolve_plot_path(
                output_root,
                plot_style,
                f"{model_name}/oneway/post",
                f"00_{model_name}_{feature}_oneway_{pred_tag}.png",
            )
            plot_oneway(
                plot_train,
                feature=feature,
                weight_col=weight_col,
                target_col="w_act",
                pred_col=pred_col,
                pred_label=pred_label,
                n_bins=n_bins,
                is_categorical=feature in oneway_categorical,
                save_path=save_path,
                show=False,
            )

    datasets = []
    if "w_act" in plot_train.columns and not plot_train["w_act"].isna().all():
        datasets.append(("Train Data", plot_train))
    if "w_act" in plot_test.columns and not plot_test["w_act"].isna().all():
        datasets.append(("Test Data", plot_test))

    def _plot_lift_for_model(pred_key: str, pred_label: str) -> None:
        if not datasets:
            return
        output_root, plot_style = _get_plot_config(pred_key)
        style = PlotStyle()
        fig, axes = plt.subplots(1, len(datasets), figsize=(11, 5))
        if len(datasets) == 1:
            axes = [axes]
        for ax, (title, data) in zip(axes, datasets):
            pred_col = f"pred_{pred_key}"
            if pred_col not in data.columns:
                continue
            plot_lift_curve(
                data[pred_col].values,
                data["w_act"].values,
                data[weight_col].values,
                n_bins=n_bins,
                title=f"Lift Chart on {title}",
                pred_label="Predicted",
                act_label="Actual",
                weight_label="Earned Exposure",
                pred_weighted=False,
                actual_weighted=True,
                ax=ax,
                show=False,
                style=style,
            )
        plt.subplots_adjust(wspace=0.3)
        filename = f"01_{model_name}_{_safe_tag(pred_label)}_lift.png"
        save_path = _resolve_plot_path(
            output_root,
            plot_style,
            f"{model_name}/lift",
            filename,
        )
        finalize_figure(fig, save_path=save_path, show=False, style=style)

    for pred_key in model_keys:
        _plot_lift_for_model(pred_key, _model_label(pred_key))

    if (
        all(k in model_keys for k in ["xgb", "resn"])
        and all(f"pred_{k}" in plot_train.columns for k in ["xgb", "resn"])
        and datasets
    ):
        style = PlotStyle()
        fig, axes = plt.subplots(1, len(datasets), figsize=(11, 5))
        if len(datasets) == 1:
            axes = [axes]
        for ax, (title, data) in zip(axes, datasets):
            plot_double_lift_curve(
                data["pred_xgb"].values,
                data["pred_resn"].values,
                data["w_act"].values,
                data[weight_col].values,
                n_bins=n_bins,
                title=f"Double Lift Chart on {title}",
                label1="Xgboost",
                label2="ResNet",
                pred1_weighted=False,
                pred2_weighted=False,
                actual_weighted=True,
                ax=ax,
                show=False,
                style=style,
            )
        plt.subplots_adjust(wspace=0.3)
        save_path = _resolve_plot_path(
            _resolve_output_dir(cfg, cfg_path),
            _resolve_plot_style(cfg),
            "",
            f"02_{model_name}_dlift_xgb_vs_resn.png",
        )
        finalize_figure(fig, save_path=save_path, show=False, style=style)

    print("Plots saved under:")
    for key in model_keys:
        output_root, _ = _get_plot_config(key)
        print(f"  - {key}: {output_root}/plot/{model_name}")
    return "Plotting complete."


def run_plot_embed(
    *,
    cfg_path: str,
    xgb_cfg_path: str,
    resn_cfg_path: str,
    ft_cfg_path: str,
    use_runtime_ft_embedding: bool = False,
) -> str:
    cfg_path = Path(cfg_path).resolve()
    xgb_cfg_path = Path(xgb_cfg_path).resolve()
    resn_cfg_path = Path(resn_cfg_path).resolve()
    ft_cfg_path = Path(ft_cfg_path).resolve()

    cfg = json.loads(cfg_path.read_text(encoding="utf-8"))
    xgb_cfg = json.loads(xgb_cfg_path.read_text(encoding="utf-8"))
    resn_cfg = json.loads(resn_cfg_path.read_text(encoding="utf-8"))
    ft_cfg = json.loads(ft_cfg_path.read_text(encoding="utf-8"))

    model_name = f"{cfg['model_list'][0]}_{cfg['model_categories'][0]}"

    raw_data_dir = (ft_cfg_path.parent / ft_cfg["data_dir"]).resolve()
    raw_path = raw_data_dir / f"{model_name}.csv"
    raw = pd.read_csv(raw_path)
    raw = _drop_duplicate_columns(raw, "raw").reset_index(drop=True)
    raw.fillna(0, inplace=True)

    ft_output_dir = (ft_cfg_path.parent / ft_cfg["output_dir"]).resolve()
    ft_prefix = ft_cfg.get("ft_feature_prefix", "ft_emb")
    raw_feature_list = _dedupe_list(ft_cfg.get("feature_list") or [])
    raw_categorical_features = _dedupe_list(ft_cfg.get("categorical_features") or [])

    if ft_cfg.get("geo_feature_nmes"):
        raise ValueError("FT inference with geo tokens is not supported in this workflow.")

    holdout_ratio = cfg.get("holdout_ratio", cfg.get("prop_test", 0.25))
    split_strategy = cfg.get("split_strategy", "random")
    split_group_col = cfg.get("split_group_col")
    split_time_col = cfg.get("split_time_col")
    split_time_ascending = cfg.get("split_time_ascending", True)
    rand_seed = cfg.get("rand_seed", 13)

    train_raw, test_raw = split_train_test(
        raw,
        holdout_ratio=holdout_ratio,
        strategy=split_strategy,
        group_col=split_group_col,
        time_col=split_time_col,
        time_ascending=split_time_ascending,
        rand_seed=rand_seed,
        reset_index_mode="none",
        ratio_label="holdout_ratio",
    )
    train_raw = _drop_duplicate_columns(train_raw, "train_raw")
    test_raw = _drop_duplicate_columns(test_raw, "test_raw")

    if use_runtime_ft_embedding:
        import torch

        ft_model_path = ft_output_dir / "model" / f"01_{model_name}_FTTransformer.pth"
        ft_payload = torch.load(ft_model_path, map_location="cpu")
        ft_model = ft_payload["model"] if isinstance(ft_payload, dict) and "model" in ft_payload else ft_payload

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        if hasattr(ft_model, "device"):
            ft_model.device = device
        if hasattr(ft_model, "to"):
            ft_model.to(device)
        if hasattr(ft_model, "ft"):
            ft_model.ft.to(device)

        emb_train = ft_model.predict(train_raw, return_embedding=True)
        emb_cols = [f"pred_{ft_prefix}_{i}" for i in range(emb_train.shape[1])]
        train_df = train_raw.copy()
        train_df[emb_cols] = emb_train

        emb_test = ft_model.predict(test_raw, return_embedding=True)
        test_df = test_raw.copy()
        test_df[emb_cols] = emb_test
    else:
        embed_data_dir = (cfg_path.parent / cfg["data_dir"]).resolve()
        embed_path = embed_data_dir / f"{model_name}.csv"
        embed_df = pd.read_csv(embed_path)
        embed_df = _drop_duplicate_columns(embed_df, "embed").reset_index(drop=True)
        embed_df.fillna(0, inplace=True)
        if len(embed_df) != len(raw):
            raise ValueError(
                f"Row count mismatch: raw={len(raw)}, embed={len(embed_df)}. "
                "Cannot align predictions to raw features."
            )
        train_df = embed_df.loc[train_raw.index].copy()
        test_df = embed_df.loc[test_raw.index].copy()

    feature_list = _dedupe_list(cfg.get("feature_list") or [])
    categorical_features = _dedupe_list(cfg.get("categorical_features") or [])

    output_dir_map = {
        "xgb": _resolve_output_dir(xgb_cfg, xgb_cfg_path),
        "resn": _resolve_output_dir(resn_cfg, resn_cfg_path),
    }
    plot_path_style_map = {
        "xgb": _resolve_plot_style(xgb_cfg),
        "resn": _resolve_plot_style(resn_cfg),
    }

    def _get_plot_config(model_key: str) -> Tuple[str, str]:
        return (
            output_dir_map.get(model_key, _resolve_output_dir(cfg, cfg_path)),
            plot_path_style_map.get(model_key, _resolve_plot_style(cfg)),
        )

    def _load_predictor(cfg_path: Path, model_key: str):
        return load_predictor_from_config(cfg_path, model_key, model_name=model_name)

    model_cfg_map = {"xgb": xgb_cfg_path, "resn": resn_cfg_path}
    model_keys = cfg.get("model_keys") or ["xgb", "resn"]
    model_keys = [key for key in model_keys if key in model_cfg_map]
    if not model_keys:
        raise ValueError("No valid model keys found in plot config.")

    default_model_labels = {"xgb": "Xgboost", "resn": "ResNet"}

    def _model_label(model_key: str) -> str:
        labels = cfg.get("model_labels") or {}
        return str(labels.get(model_key, default_model_labels.get(model_key, model_key)))

    predictors = {key: _load_predictor(model_cfg_map[key], key) for key in model_keys}
    pred_train = {}
    pred_test = {}
    for key, predictor in predictors.items():
        pred_train[key] = predictor.predict(train_df).reshape(-1)
        pred_test[key] = predictor.predict(test_df).reshape(-1)
        if len(pred_train[key]) != len(train_df):
            raise ValueError(f"Train prediction length mismatch for {key}")
        if len(pred_test[key]) != len(test_df):
            raise ValueError(f"Test prediction length mismatch for {key}")

    plot_train = train_raw.copy()
    plot_test = test_raw.copy()
    for key in model_keys:
        plot_train[f"pred_{key}"] = pred_train[key]
        plot_test[f"pred_{key}"] = pred_test[key]

    weight_col = cfg["weight"]
    target_col = cfg["target"]
    if weight_col not in plot_train.columns:
        plot_train[weight_col] = 1.0
    if weight_col not in plot_test.columns:
        plot_test[weight_col] = 1.0
    if target_col in plot_train.columns:
        plot_train["w_act"] = plot_train[target_col] * plot_train[weight_col]
    if target_col in plot_test.columns:
        plot_test["w_act"] = plot_test[target_col] * plot_test[weight_col]

    if "w_act" not in plot_train.columns or plot_train["w_act"].isna().all():
        print("[Plot] Missing target values in train split; skip plots.")
        return "Skipped plotting due to missing target values."

    n_bins = cfg.get("plot", {}).get("n_bins", 10)
    oneway_features = raw_feature_list or feature_list
    oneway_categorical = set(raw_categorical_features or categorical_features)

    for pred_key in model_keys:
        pred_label = _model_label(pred_key)
        pred_col = f"pred_{pred_key}"
        pred_tag = _safe_tag(pred_label or pred_col)
        output_root, plot_style = _get_plot_config(pred_key)
        for feature in oneway_features:
            if feature not in plot_train.columns:
                continue
            save_path = _resolve_plot_path(
                output_root,
                plot_style,
                f"{model_name}/oneway/post",
                f"00_{model_name}_{feature}_oneway_{pred_tag}.png",
            )
            plot_oneway(
                plot_train,
                feature=feature,
                weight_col=weight_col,
                target_col="w_act",
                pred_col=pred_col,
                pred_label=pred_label,
                n_bins=n_bins,
                is_categorical=feature in oneway_categorical,
                save_path=save_path,
                show=False,
            )

    datasets = []
    if "w_act" in plot_train.columns and not plot_train["w_act"].isna().all():
        datasets.append(("Train Data", plot_train))
    if "w_act" in plot_test.columns and not plot_test["w_act"].isna().all():
        datasets.append(("Test Data", plot_test))

    def _plot_lift_for_model(pred_key: str, pred_label: str) -> None:
        if not datasets:
            return
        output_root, plot_style = _get_plot_config(pred_key)
        style = PlotStyle()
        fig, axes = plt.subplots(1, len(datasets), figsize=(11, 5))
        if len(datasets) == 1:
            axes = [axes]
        for ax, (title, data) in zip(axes, datasets):
            pred_col = f"pred_{pred_key}"
            if pred_col not in data.columns:
                continue
            plot_lift_curve(
                data[pred_col].values,
                data["w_act"].values,
                data[weight_col].values,
                n_bins=n_bins,
                title=f"Lift Chart on {title}",
                pred_label="Predicted",
                act_label="Actual",
                weight_label="Earned Exposure",
                pred_weighted=False,
                actual_weighted=True,
                ax=ax,
                show=False,
                style=style,
            )
        plt.subplots_adjust(wspace=0.3)
        filename = f"01_{model_name}_{_safe_tag(pred_label)}_lift.png"
        save_path = _resolve_plot_path(
            output_root,
            plot_style,
            f"{model_name}/lift",
            filename,
        )
        finalize_figure(fig, save_path=save_path, show=False, style=style)

    for pred_key in model_keys:
        _plot_lift_for_model(pred_key, _model_label(pred_key))

    if (
        all(k in model_keys for k in ["xgb", "resn"])
        and all(f"pred_{k}" in plot_train.columns for k in ["xgb", "resn"])
        and datasets
    ):
        style = PlotStyle()
        fig, axes = plt.subplots(1, len(datasets), figsize=(11, 5))
        if len(datasets) == 1:
            axes = [axes]
        for ax, (title, data) in zip(axes, datasets):
            plot_double_lift_curve(
                data["pred_xgb"].values,
                data["pred_resn"].values,
                data["w_act"].values,
                data[weight_col].values,
                n_bins=n_bins,
                title=f"Double Lift Chart on {title}",
                label1="Xgboost",
                label2="ResNet",
                pred1_weighted=False,
                pred2_weighted=False,
                actual_weighted=True,
                ax=ax,
                show=False,
                style=style,
            )
        plt.subplots_adjust(wspace=0.3)
        save_path = _resolve_plot_path(
            _resolve_output_dir(cfg, cfg_path),
            _resolve_plot_style(cfg),
            "",
            f"02_{model_name}_dlift_xgb_vs_resn.png",
        )
        finalize_figure(fig, save_path=save_path, show=False, style=style)

    print("Plots saved under:")
    for key in model_keys:
        output_root, _ = _get_plot_config(key)
        print(f"  - {key}: {output_root}/plot/{model_name}")
    return "Plotting complete."




"""CLI common utilities.

This module re-exports shared utilities from ins_pricing.utils and provides
CLI-specific functionality for configuration loading and train/test splitting.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Tuple

import pandas as pd
from sklearn.model_selection import GroupShuffleSplit, train_test_split

# Re-export shared utilities for backward compatibility
from ins_pricing.utils.paths import (
    PLOT_MODEL_LABELS,
    PYTORCH_TRAINERS,
    dedupe_preserve_order,
    build_model_names,
    parse_model_pairs,
    resolve_path,
    resolve_dir_path,
    resolve_data_path,
    load_dataset,
    coerce_dataset_types,
    fingerprint_file,
)

__all__ = [
    # From shared utils
    "PLOT_MODEL_LABELS",
    "PYTORCH_TRAINERS",
    "dedupe_preserve_order",
    "build_model_names",
    "parse_model_pairs",
    "resolve_path",
    "resolve_dir_path",
    "resolve_data_path",
    "load_dataset",
    "coerce_dataset_types",
    "fingerprint_file",
    # CLI-specific
    "split_train_test",
    "resolve_config_path",
    "load_config_json",
    "set_env",
    "normalize_config_paths",
    "resolve_dtype_map",
    "resolve_data_config",
    "resolve_report_config",
    "resolve_split_config",
    "resolve_runtime_config",
    "resolve_output_dirs",
    "resolve_and_load_config",
]


# =============================================================================
# CLI-specific: Train/Test Splitting
# =============================================================================


def split_train_test(
    df: pd.DataFrame,
    *,
    holdout_ratio: float,
    strategy: str = "random",
    group_col: Optional[str] = None,
    time_col: Optional[str] = None,
    time_ascending: bool = True,
    rand_seed: Optional[int] = None,
    reset_index_mode: str = "none",
    ratio_label: str = "holdout_ratio",
    include_strategy_in_ratio_error: bool = False,
    validate_ratio: bool = True,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Split a DataFrame into train and test sets.

    Args:
        df: Input DataFrame
        holdout_ratio: Proportion for test set (0.0-1.0)
        strategy: Split strategy ('random', 'time', 'group')
        group_col: Column name for group-based splitting
        time_col: Column name for time-based splitting
        time_ascending: Sort order for time-based splitting
        rand_seed: Random seed for reproducibility
        reset_index_mode: When to reset index ('none', 'always', 'time_group')
        ratio_label: Label for ratio in error messages
        include_strategy_in_ratio_error: Include strategy in error messages
        validate_ratio: Whether to validate ratio bounds

    Returns:
        Tuple of (train_df, test_df)
    """
    strategy = str(strategy or "random").strip().lower()
    holdout_ratio = float(holdout_ratio)

    if include_strategy_in_ratio_error and strategy in {
        "time",
        "timeseries",
        "temporal",
        "group",
        "grouped",
    }:
        strategy_label = (
            "time" if strategy in {"time", "timeseries", "temporal"} else "group"
        )
        ratio_error = (
            f"{ratio_label} must be in (0, 1) for {strategy_label} split; "
            f"got {holdout_ratio}."
        )
    else:
        ratio_error = f"{ratio_label} must be in (0, 1); got {holdout_ratio}."

    if strategy in {"time", "timeseries", "temporal"}:
        if not time_col:
            raise ValueError("split_time_col is required for time split_strategy.")
        if time_col not in df.columns:
            raise KeyError(f"split_time_col '{time_col}' not in dataset columns.")
        if validate_ratio and not (0.0 < holdout_ratio < 1.0):
            raise ValueError(ratio_error)
        ordered = df.sort_values(time_col, ascending=bool(time_ascending))
        cutoff = int(len(ordered) * (1.0 - holdout_ratio))
        if cutoff <= 0 or cutoff >= len(ordered):
            raise ValueError(
                f"{ratio_label}={holdout_ratio} leaves no data for train/test split."
            )
        train_df = ordered.iloc[:cutoff]
        test_df = ordered.iloc[cutoff:]
    elif strategy in {"group", "grouped"}:
        if not group_col:
            raise ValueError("split_group_col is required for group split_strategy.")
        if group_col not in df.columns:
            raise KeyError(f"split_group_col '{group_col}' not in dataset columns.")
        if validate_ratio and not (0.0 < holdout_ratio < 1.0):
            raise ValueError(ratio_error)
        splitter = GroupShuffleSplit(
            n_splits=1,
            test_size=holdout_ratio,
            random_state=rand_seed,
        )
        train_idx, test_idx = next(splitter.split(df, groups=df[group_col]))
        train_df = df.iloc[train_idx]
        test_df = df.iloc[test_idx]
    else:
        train_df, test_df = train_test_split(
            df, test_size=holdout_ratio, random_state=rand_seed
        )

    if reset_index_mode == "always" or (
        reset_index_mode == "time_group"
        and strategy in {"time", "timeseries", "temporal", "group", "grouped"}
    ):
        train_df = train_df.reset_index(drop=True)
        test_df = test_df.reset_index(drop=True)

    return train_df, test_df


# =============================================================================
# CLI-specific: Configuration Loading (delegated to cli_config)
# =============================================================================


def _load_cli_config():
    """Load the cli_config module."""
    try:
        from ins_pricing.cli.utils import cli_config as _cli_config
    except Exception:
        import cli_config as _cli_config
    return _cli_config


def resolve_config_path(raw: str, script_dir: Path) -> Path:
    """Resolve a configuration file path."""
    return _load_cli_config().resolve_config_path(raw, script_dir)


def load_config_json(path: Path, required_keys: Sequence[str]) -> Dict[str, Any]:
    """Load and validate a JSON configuration file."""
    return _load_cli_config().load_config_json(path, required_keys)


def set_env(env_overrides: Dict[str, Any]) -> None:
    """Set environment variables from configuration."""
    _load_cli_config().set_env(env_overrides)


def normalize_config_paths(cfg: Dict[str, Any], config_path: Path) -> Dict[str, Any]:
    """Normalize paths in configuration relative to config file location."""
    return _load_cli_config().normalize_config_paths(cfg, config_path)


def resolve_dtype_map(value: Any, base_dir: Path) -> Dict[str, Any]:
    """Resolve dtype mapping from configuration."""
    return _load_cli_config().resolve_dtype_map(value, base_dir)


def resolve_data_config(
    cfg: Dict[str, Any],
    config_path: Path,
    *,
    create_data_dir: bool = False,
) -> Tuple[Path, str, Optional[str], Dict[str, Any]]:
    """Resolve data configuration from config file."""
    return _load_cli_config().resolve_data_config(
        cfg,
        config_path,
        create_data_dir=create_data_dir,
    )


def resolve_report_config(cfg: Dict[str, Any]) -> Dict[str, Any]:
    """Resolve report configuration."""
    return _load_cli_config().resolve_report_config(cfg)


def resolve_split_config(cfg: Dict[str, Any]) -> Dict[str, Any]:
    """Resolve train/test split configuration."""
    return _load_cli_config().resolve_split_config(cfg)


def resolve_runtime_config(cfg: Dict[str, Any]) -> Dict[str, Any]:
    """Resolve runtime configuration."""
    return _load_cli_config().resolve_runtime_config(cfg)


def resolve_output_dirs(
    cfg: Dict[str, Any],
    config_path: Path,
    *,
    output_override: Optional[str] = None,
) -> Dict[str, Optional[str]]:
    """Resolve output directory paths."""
    return _load_cli_config().resolve_output_dirs(
        cfg,
        config_path,
        output_override=output_override,
    )


def resolve_and_load_config(
    raw: str,
    script_dir: Path,
    required_keys: Sequence[str],
    *,
    apply_env: bool = True,
) -> Tuple[Path, Dict[str, Any]]:
    """Resolve and load a configuration file."""
    return _load_cli_config().resolve_and_load_config(
        raw,
        script_dir,
        required_keys,
        apply_env=apply_env,
    )



from __future__ import annotations

from typing import Iterable, Optional, Tuple

import numpy as np
import pandas as pd
from sklearn.model_selection import (
    GroupKFold,
    GroupShuffleSplit,
    KFold,
    ShuffleSplit,
    TimeSeriesSplit,
)


class _OrderSplitter:
    def __init__(self, splitter, order: np.ndarray) -> None:
        self._splitter = splitter
        self._order = np.asarray(order)

    def split(self, X, y=None, groups=None):
        order = self._order
        X_ord = X.iloc[order] if hasattr(X, "iloc") else X[order]
        for tr_idx, val_idx in self._splitter.split(X_ord, y=y, groups=groups):
            yield order[tr_idx], order[val_idx]


class CVStrategyResolver:
    """Resolve CV splitters for random/time/group strategies."""

    TIME_STRATEGIES = {"time", "timeseries", "temporal"}
    GROUP_STRATEGIES = {"group", "grouped"}

    def __init__(self, config, train_data: pd.DataFrame, rand_seed: Optional[int] = None):
        self.config = config
        self.train_data = train_data
        self.rand_seed = rand_seed
        self._strategy = self._normalize_strategy()

    def _normalize_strategy(self) -> str:
        raw = str(getattr(self.config, "cv_strategy", "random") or "random")
        return raw.strip().lower()

    @property
    def strategy(self) -> str:
        return self._strategy

    def is_time_strategy(self) -> bool:
        return self._strategy in self.TIME_STRATEGIES

    def is_group_strategy(self) -> bool:
        return self._strategy in self.GROUP_STRATEGIES

    def get_time_col(self) -> str:
        time_col = getattr(self.config, "cv_time_col", None)
        if not time_col:
            raise ValueError("cv_time_col is required for time cv_strategy.")
        if time_col not in self.train_data.columns:
            raise KeyError(f"cv_time_col '{time_col}' not in train_data.")
        return time_col

    def get_time_ascending(self) -> bool:
        return bool(getattr(self.config, "cv_time_ascending", True))

    def get_group_col(self) -> str:
        group_col = getattr(self.config, "cv_group_col", None)
        if not group_col:
            raise ValueError("cv_group_col is required for group cv_strategy.")
        if group_col not in self.train_data.columns:
            raise KeyError(f"cv_group_col '{group_col}' not in train_data.")
        return group_col

    def get_time_ordered_indices(self, X_all: pd.DataFrame) -> np.ndarray:
        time_col = self.get_time_col()
        ascending = self.get_time_ascending()
        order_index = self.train_data[time_col].sort_values(ascending=ascending).index
        index_set = set(X_all.index)
        order_index = [idx for idx in order_index if idx in index_set]
        order = X_all.index.get_indexer(order_index)
        return order[order >= 0]

    def get_groups(self, X_all: pd.DataFrame) -> pd.Series:
        group_col = self.get_group_col()
        return self.train_data.reindex(X_all.index)[group_col]

    def create_train_val_splitter(
        self,
        X_all: pd.DataFrame,
        val_ratio: float,
    ) -> Tuple[Optional[Tuple[np.ndarray, np.ndarray]], Optional[pd.Series]]:
        if self.is_time_strategy():
            order = self.get_time_ordered_indices(X_all)
            cutoff = int(len(order) * (1.0 - val_ratio))
            if cutoff <= 0 or cutoff >= len(order):
                raise ValueError(f"val_ratio={val_ratio} leaves no data for train/val split.")
            return (order[:cutoff], order[cutoff:]), None

        if self.is_group_strategy():
            groups = self.get_groups(X_all)
            splitter = GroupShuffleSplit(
                n_splits=1, test_size=val_ratio, random_state=self.rand_seed
            )
            train_idx, val_idx = next(splitter.split(X_all, groups=groups))
            return (train_idx, val_idx), groups

        splitter = ShuffleSplit(
            n_splits=1, test_size=val_ratio, random_state=self.rand_seed
        )
        train_idx, val_idx = next(splitter.split(X_all))
        return (train_idx, val_idx), None

    def create_cv_splitter(
        self,
        X_all: pd.DataFrame,
        y_all: Optional[pd.Series],
        n_splits: int,
        val_ratio: float,
    ) -> Tuple[Iterable[Tuple[np.ndarray, np.ndarray]], int]:
        n_splits = max(2, int(n_splits))

        if self.is_group_strategy():
            groups = self.get_groups(X_all)
            n_groups = int(groups.nunique(dropna=False))
            if n_groups < 2:
                return iter([]), 0
            n_splits = min(n_splits, n_groups)
            if n_splits < 2:
                return iter([]), 0
            splitter = GroupKFold(n_splits=n_splits)
            return splitter.split(X_all, y_all, groups=groups), n_splits

        if self.is_time_strategy():
            order = self.get_time_ordered_indices(X_all)
            if len(order) < 2:
                return iter([]), 0
            n_splits = min(n_splits, max(2, len(order) - 1))
            if n_splits < 2:
                return iter([]), 0
            splitter = TimeSeriesSplit(n_splits=n_splits)
            return _OrderSplitter(splitter, order).split(X_all), n_splits

        if len(X_all) < n_splits:
            n_splits = len(X_all)
        if n_splits < 2:
            return iter([]), 0
        splitter = ShuffleSplit(
            n_splits=n_splits, test_size=val_ratio, random_state=self.rand_seed
        )
        return splitter.split(X_all), n_splits

    def create_kfold_splitter(
        self,
        X_all: pd.DataFrame,
        k: int,
    ) -> Tuple[Optional[Iterable[Tuple[np.ndarray, np.ndarray]]], int]:
        k = max(2, int(k))
        n_samples = len(X_all)
        if n_samples < 2:
            return None, 0

        if self.is_group_strategy():
            groups = self.get_groups(X_all)
            n_groups = int(groups.nunique(dropna=False))
            if n_groups < 2:
                return None, 0
            k = min(k, n_groups)
            if k < 2:
                return None, 0
            splitter = GroupKFold(n_splits=k)
            return splitter.split(X_all, y=None, groups=groups), k

        if self.is_time_strategy():
            order = self.get_time_ordered_indices(X_all)
            if len(order) < 2:
                return None, 0
            k = min(k, max(2, len(order) - 1))
            if k < 2:
                return None, 0
            splitter = TimeSeriesSplit(n_splits=k)
            return _OrderSplitter(splitter, order).split(X_all), k

        k = min(k, n_samples)
        if k < 2:
            return None, 0
        splitter = KFold(n_splits=k, shuffle=True, random_state=self.rand_seed)
        return splitter.split(X_all), k


__all__ = ["CVStrategyResolver", "_OrderSplitter"]



"""Tests for production scoring module."""

import numpy as np
import pandas as pd
import pytest
from unittest.mock import Mock, patch

from ins_pricing.exceptions import DataValidationError


@pytest.fixture
def sample_predictions():
    """Sample prediction data."""
    return pd.DataFrame({
        "actual": [100, 150, 200, 250, 300],
        "predicted": [105, 145, 210, 240, 295],
        "weight": [1.0, 1.0, 1.0, 1.0, 1.0]
    })


@pytest.fixture
def classification_data():
    """Sample classification data."""
    return pd.DataFrame({
        "actual": [0, 1, 1, 0, 1, 0, 1, 1, 0, 0],
        "predicted_proba": [0.1, 0.9, 0.8, 0.2, 0.7, 0.3, 0.85, 0.6, 0.15, 0.25],
        "predicted_class": [0, 1, 1, 0, 1, 0, 1, 1, 0, 0]
    })


class TestRegressionMetrics:
    """Test regression scoring metrics."""

    def test_weighted_mse(self, sample_predictions):
        """Test weighted mean squared error calculation."""
        from ins_pricing.production.scoring import weighted_mse

        mse = weighted_mse(
            sample_predictions['actual'],
            sample_predictions['predicted'],
            sample_predictions['weight']
        )

        assert isinstance(mse, (int, float, np.number))
        assert mse >= 0

    def test_weighted_mae(self, sample_predictions):
        """Test weighted mean absolute error calculation."""
        from ins_pricing.production.scoring import weighted_mae

        mae = weighted_mae(
            sample_predictions['actual'],
            sample_predictions['predicted'],
            sample_predictions['weight']
        )

        assert isinstance(mae, (int, float, np.number))
        assert mae >= 0

    def test_weighted_r2(self, sample_predictions):
        """Test weighted R² score calculation."""
        from ins_pricing.production.scoring import weighted_r2

        r2 = weighted_r2(
            sample_predictions['actual'],
            sample_predictions['predicted'],
            sample_predictions['weight']
        )

        assert isinstance(r2, (int, float, np.number))
        assert r2 <= 1.0

    def test_mape(self, sample_predictions):
        """Test mean absolute percentage error."""
        from ins_pricing.production.scoring import mape

        mape_score = mape(
            sample_predictions['actual'],
            sample_predictions['predicted']
        )

        assert isinstance(mape_score, (int, float, np.number))
        assert mape_score >= 0

    def test_metrics_with_zero_actuals(self):
        """Test metrics handling when actual values are zero."""
        from ins_pricing.production.scoring import mape

        data = pd.DataFrame({
            "actual": [0, 100, 200],
            "predicted": [10, 105, 195]
        })

        # MAPE should handle zeros gracefully
        with pytest.raises((ValueError, ZeroDivisionError)):
            mape(data['actual'], data['predicted'])


class TestClassificationMetrics:
    """Test classification scoring metrics."""

    def test_accuracy(self, classification_data):
        """Test accuracy calculation."""
        from ins_pricing.production.scoring import accuracy

        acc = accuracy(
            classification_data['actual'],
            classification_data['predicted_class']
        )

        assert 0 <= acc <= 1

    def test_precision_recall(self, classification_data):
        """Test precision and recall calculation."""
        from ins_pricing.production.scoring import precision_recall

        precision, recall = precision_recall(
            classification_data['actual'],
            classification_data['predicted_class']
        )

        assert 0 <= precision <= 1
        assert 0 <= recall <= 1

    def test_f1_score(self, classification_data):
        """Test F1 score calculation."""
        from ins_pricing.production.scoring import f1_score

        f1 = f1_score(
            classification_data['actual'],
            classification_data['predicted_class']
        )

        assert 0 <= f1 <= 1

    def test_roc_auc(self, classification_data):
        """Test ROC AUC calculation."""
        from ins_pricing.production.scoring import roc_auc

        auc = roc_auc(
            classification_data['actual'],
            classification_data['predicted_proba']
        )

        assert 0 <= auc <= 1

    def test_confusion_matrix(self, classification_data):
        """Test confusion matrix generation."""
        from ins_pricing.production.scoring import confusion_matrix

        cm = confusion_matrix(
            classification_data['actual'],
            classification_data['predicted_class']
        )

        assert cm.shape == (2, 2)
        assert np.all(cm >= 0)


class TestInsuranceMetrics:
    """Test insurance-specific metrics."""

    def test_loss_ratio(self):
        """Test loss ratio calculation."""
        from ins_pricing.production.scoring import loss_ratio

        data = pd.DataFrame({
            "claims": [100, 200, 150],
            "premiums": [120, 180, 160],
            "exposure": [1.0, 1.0, 1.0]
        })

        lr = loss_ratio(data['claims'], data['premiums'], data['exposure'])

        assert isinstance(lr, (int, float, np.number))
        assert lr >= 0

    def test_gini_coefficient(self, sample_predictions):
        """Test Gini coefficient calculation."""
        from ins_pricing.production.scoring import gini_coefficient

        gini = gini_coefficient(
            sample_predictions['actual'],
            sample_predictions['predicted']
        )

        assert -1 <= gini <= 1

    def test_lift_at_percentile(self, sample_predictions):
        """Test lift calculation at specific percentile."""
        from ins_pricing.production.scoring import lift_at_percentile

        lift = lift_at_percentile(
            sample_predictions['actual'],
            sample_predictions['predicted'],
            percentile=20
        )

        assert isinstance(lift, (int, float, np.number))


class TestMetricValidation:
    """Test metric input validation."""

    def test_mismatched_lengths(self):
        """Test error on mismatched array lengths."""
        from ins_pricing.production.scoring import weighted_mse
        from ins_pricing.utils.validation import validate_dataframe_not_empty

        actual = np.array([1, 2, 3])
        predicted = np.array([1, 2])  # Wrong length
        weights = np.array([1, 1, 1])

        with pytest.raises((ValueError, IndexError)):
            weighted_mse(actual, predicted, weights)

    def test_negative_weights(self):
        """Test handling of negative weights."""
        from ins_pricing.production.scoring import weighted_mse

        actual = np.array([100, 200, 300])
        predicted = np.array([105, 195, 310])
        weights = np.array([1.0, -1.0, 1.0])  # Negative weight

        with pytest.raises(ValueError):
            weighted_mse(actual, predicted, weights)

    def test_nan_values(self):
        """Test handling of NaN values."""
        from ins_pricing.production.scoring import weighted_mse

        actual = np.array([100, np.nan, 300])
        predicted = np.array([105, 195, 310])
        weights = np.array([1.0, 1.0, 1.0])

        with pytest.raises(ValueError):
            weighted_mse(actual, predicted, weights)


class TestScoringReport:
    """Test scoring report generation."""

    def test_generate_regression_report(self, sample_predictions):
        """Test comprehensive regression scoring report."""
        from ins_pricing.production.scoring import generate_scoring_report

        report = generate_scoring_report(
            actual=sample_predictions['actual'],
            predicted=sample_predictions['predicted'],
            weights=sample_predictions['weight'],
            task_type='regression'
        )

        assert 'mse' in report
        assert 'mae' in report
        assert 'r2' in report
        assert all(isinstance(v, (int, float, np.number)) for v in report.values())

    def test_generate_classification_report(self, classification_data):
        """Test comprehensive classification scoring report."""
        from ins_pricing.production.scoring import generate_scoring_report

        report = generate_scoring_report(
            actual=classification_data['actual'],
            predicted=classification_data['predicted_class'],
            predicted_proba=classification_data['predicted_proba'],
            task_type='classification'
        )

        assert 'accuracy' in report
        assert 'precision' in report
        assert 'recall' in report
        assert 'f1' in report
        assert 'roc_auc' in report

    def test_save_report_to_file(self, sample_predictions, tmp_path):
        """Test saving scoring report to file."""
        from ins_pricing.production.scoring import generate_scoring_report, save_report

        report = generate_scoring_report(
            actual=sample_predictions['actual'],
            predicted=sample_predictions['predicted'],
            task_type='regression'
        )

        output_path = tmp_path / "scoring_report.json"
        save_report(report, output_path)

        assert output_path.exists()


@pytest.mark.performance
class TestScoringPerformance:
    """Test scoring performance on large datasets."""

    def test_large_dataset_scoring(self):
        """Test scoring metrics on large dataset."""
        from ins_pricing.production.scoring import weighted_mse

        n = 1_000_000
        actual = np.random.uniform(100, 500, n)
        predicted = actual + np.random.normal(0, 20, n)
        weights = np.ones(n)

        import time
        start = time.time()
        mse = weighted_mse(actual, predicted, weights)
        elapsed = time.time() - start

        assert isinstance(mse, (int, float, np.number))
        assert elapsed < 1.0  # Should complete in under 1 second



"""Torch compatibility helpers for 1.x and 2.x support."""

from __future__ import annotations

import inspect
import os
from typing import Any, Optional

try:
    import torch

    TORCH_AVAILABLE = True
except ImportError:  # pragma: no cover - handled by callers
    TORCH_AVAILABLE = False
    torch = None

_SUPPORTS_WEIGHTS_ONLY: Optional[bool] = None
_DYNAMO_PATCHED = False


def _supports_weights_only() -> bool:
    """Check whether torch.load supports the weights_only argument."""
    global _SUPPORTS_WEIGHTS_ONLY
    if _SUPPORTS_WEIGHTS_ONLY is None:
        if not TORCH_AVAILABLE:
            _SUPPORTS_WEIGHTS_ONLY = False
        else:
            try:
                sig = inspect.signature(torch.load)
                _SUPPORTS_WEIGHTS_ONLY = "weights_only" in sig.parameters
            except (TypeError, ValueError):
                _SUPPORTS_WEIGHTS_ONLY = False
    return bool(_SUPPORTS_WEIGHTS_ONLY)


def torch_load(
    path: Any,
    *args: Any,
    weights_only: Optional[bool] = None,
    **kwargs: Any,
) -> Any:
    """Load a torch artifact while handling 1.x/2.x API differences."""
    if not TORCH_AVAILABLE:
        raise RuntimeError("torch is required to load model files.")
    if weights_only is not None and _supports_weights_only():
        return torch.load(path, *args, weights_only=weights_only, **kwargs)
    return torch.load(path, *args, **kwargs)


def _env_truthy(key: str) -> bool:
    value = os.environ.get(key)
    if value is None:
        return False
    return str(value).strip().lower() in {"1", "true", "yes", "y", "on"}


def disable_torch_dynamo_if_requested() -> None:
    """Disable torch._dynamo wrappers when compile is explicitly disabled."""
    global _DYNAMO_PATCHED
    if _DYNAMO_PATCHED or not TORCH_AVAILABLE:
        return

    if not any(
        _env_truthy(k)
        for k in (
            "TORCHDYNAMO_DISABLE",
            "TORCH_DISABLE_DYNAMO",
            "TORCH_COMPILE_DISABLE",
            "TORCHINDUCTOR_DISABLE",
        )
    ):
        return

    try:
        import torch.optim.optimizer as optim_mod
    except Exception:
        return

    for name in ("state_dict", "load_state_dict", "zero_grad", "add_param_group"):
        fn = getattr(optim_mod.Optimizer, name, None)
        wrapped = getattr(fn, "__wrapped__", None)
        if wrapped is not None:
            setattr(optim_mod.Optimizer, name, wrapped)

    _DYNAMO_PATCHED = True



"""Prediction workflows used by the frontend UI."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Optional

import pandas as pd

from ins_pricing.production.inference import load_predictor_from_config


def run_predict_ft_embed(
    *,
    ft_cfg_path: str,
    xgb_cfg_path: Optional[str],
    resn_cfg_path: Optional[str],
    input_path: str,
    output_path: str,
    model_name: Optional[str],
    model_keys: str,
) -> str:
    ft_cfg_path = Path(ft_cfg_path).resolve()
    xgb_cfg_path = Path(xgb_cfg_path).resolve() if xgb_cfg_path else None
    resn_cfg_path = Path(resn_cfg_path).resolve() if resn_cfg_path else None
    input_path = Path(input_path).resolve()
    output_path = Path(output_path).resolve()

    if not input_path.exists():
        raise FileNotFoundError(f"Input data not found: {input_path}")

    keys = [k.strip() for k in model_keys.split(",") if k.strip()]
    if not keys:
        raise ValueError("model_keys is empty.")

    ft_cfg = json.loads(ft_cfg_path.read_text(encoding="utf-8"))
    xgb_cfg = json.loads(xgb_cfg_path.read_text(encoding="utf-8")) if xgb_cfg_path else None
    resn_cfg = json.loads(resn_cfg_path.read_text(encoding="utf-8")) if resn_cfg_path else None

    if model_name is None:
        model_list = list(ft_cfg.get("model_list") or [])
        model_categories = list(ft_cfg.get("model_categories") or [])
        if len(model_list) != 1 or len(model_categories) != 1:
            raise ValueError("Set model_name when multiple models exist.")
        model_name = f"{model_list[0]}_{model_categories[0]}"

    ft_output_dir = (ft_cfg_path.parent / ft_cfg["output_dir"]).resolve()
    xgb_output_dir = (xgb_cfg_path.parent / xgb_cfg["output_dir"]).resolve() if xgb_cfg else None
    ft_prefix = ft_cfg.get("ft_feature_prefix", "ft_emb")
    xgb_task_type = str(xgb_cfg.get("task_type", "regression")) if xgb_cfg else None

    if ft_cfg.get("geo_feature_nmes"):
        raise ValueError("FT with geo tokens is not supported in this workflow.")

    import torch
    import joblib

    print("Loading FT model...")
    ft_model_path = ft_output_dir / "model" / f"01_{model_name}_FTTransformer.pth"
    ft_payload = torch.load(ft_model_path, map_location="cpu")
    ft_model = ft_payload["model"] if isinstance(ft_payload, dict) and "model" in ft_payload else ft_payload

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if hasattr(ft_model, "device"):
        ft_model.device = device
    if hasattr(ft_model, "to"):
        ft_model.to(device)
    if hasattr(ft_model, "ft"):
        ft_model.ft.to(device)

    df_new = pd.read_csv(input_path)
    emb = ft_model.predict(df_new, return_embedding=True)
    emb_cols = [f"pred_{ft_prefix}_{i}" for i in range(emb.shape[1])]
    df_with_emb = df_new.copy()
    df_with_emb[emb_cols] = emb
    result = df_with_emb.copy()

    if "xgb" in keys:
        if not xgb_cfg or not xgb_output_dir:
            raise ValueError("xgb model selected but xgb_cfg_path is missing.")
        xgb_model_path = xgb_output_dir / "model" / f"01_{model_name}_Xgboost.pkl"
        xgb_payload = joblib.load(xgb_model_path)
        if isinstance(xgb_payload, dict) and "model" in xgb_payload:
            xgb_model = xgb_payload["model"]
            feature_list = xgb_payload.get("preprocess_artifacts", {}).get("factor_nmes")
        else:
            xgb_model = xgb_payload
            feature_list = None
        if not feature_list:
            feature_list = xgb_cfg.get("feature_list") or []
        if not feature_list:
            raise ValueError("Feature list missing for XGB model.")

        X = df_with_emb[feature_list]
        if xgb_task_type == "classification" and hasattr(xgb_model, "predict_proba"):
            pred = xgb_model.predict_proba(X)[:, 1]
        else:
            pred = xgb_model.predict(X)
        result["pred_xgb"] = pred

    if "resn" in keys:
        if not resn_cfg_path:
            raise ValueError("resn model selected but resn_cfg_path is missing.")
        resn_predictor = load_predictor_from_config(
            resn_cfg_path, "resn", model_name=model_name
        )
        pred_resn = resn_predictor.predict(df_with_emb)
        result["pred_resn"] = pred_resn

    output_path.parent.mkdir(parents=True, exist_ok=True)
    result.to_csv(output_path, index=False)
    print(f"Saved predictions to: {output_path}")
    return str(output_path)




from __future__ import annotations

import argparse
import os
from pathlib import Path
from typing import Any, Dict, List, Mapping, Tuple

import matplotlib

if os.name != "nt" and not os.environ.get("DISPLAY") and not os.environ.get("MPLBACKEND"):
    matplotlib.use("Agg")
import matplotlib.pyplot as plt


def parse_bayesopt_args(
    add_config_json_arg,
    add_output_dir_arg,
) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Batch trainer generated from BayesOpt_AutoPricing notebook."
    )
    add_config_json_arg(
        parser,
        help_text="Path to the JSON config describing datasets and feature columns.",
    )
    parser.add_argument(
        "--model-keys",
        nargs="+",
        default=["ft"],
        choices=["glm", "xgb", "resn", "ft", "gnn", "all"],
        help="Space-separated list of trainers to run (e.g., --model-keys glm xgb). Include 'all' to run every trainer.",
    )
    parser.add_argument(
        "--stack-model-keys",
        nargs="+",
        default=None,
        choices=["glm", "xgb", "resn", "ft", "gnn", "all"],
        help=(
            "Only used when ft_role != 'model' (FT runs as feature generator). "
            "When provided (or when config defines stack_model_keys), these trainers run after FT features "
            "are generated. Use 'all' to run every non-FT trainer."
        ),
    )
    parser.add_argument(
        "--max-evals",
        type=int,
        default=50,
        help="Optuna trial count per dataset.",
    )
    parser.add_argument(
        "--use-resn-ddp",
        action="store_true",
        help="Force ResNet trainer to use DistributedDataParallel.",
    )
    parser.add_argument(
        "--use-ft-ddp",
        action="store_true",
        help="Force FT-Transformer trainer to use DistributedDataParallel.",
    )
    parser.add_argument(
        "--use-resn-dp",
        action="store_true",
        help="Enable ResNet DataParallel fall-back regardless of config.",
    )
    parser.add_argument(
        "--use-ft-dp",
        action="store_true",
        help="Enable FT-Transformer DataParallel fall-back regardless of config.",
    )
    parser.add_argument(
        "--use-gnn-dp",
        action="store_true",
        help="Enable GNN DataParallel fall-back regardless of config.",
    )
    parser.add_argument(
        "--gnn-no-ann",
        action="store_true",
        help="Disable approximate k-NN for GNN graph construction and use exact search.",
    )
    parser.add_argument(
        "--gnn-ann-threshold",
        type=int,
        default=None,
        help="Row threshold above which approximate k-NN is preferred (overrides config).",
    )
    parser.add_argument(
        "--gnn-graph-cache",
        default=None,
        help="Optional path to persist/load cached adjacency matrix for GNN.",
    )
    parser.add_argument(
        "--gnn-max-gpu-nodes",
        type=int,
        default=None,
        help="Overrides the maximum node count allowed for GPU k-NN graph construction.",
    )
    parser.add_argument(
        "--gnn-gpu-mem-ratio",
        type=float,
        default=None,
        help="Overrides the fraction of free GPU memory the k-NN builder may consume.",
    )
    parser.add_argument(
        "--gnn-gpu-mem-overhead",
        type=float,
        default=None,
        help="Overrides the temporary GPU memory overhead multiplier for k-NN estimation.",
    )
    add_output_dir_arg(
        parser,
        help_text="Override output root for models/results/plots.",
    )
    parser.add_argument(
        "--plot-curves",
        action="store_true",
        help="Enable lift/diagnostic plots after training (config file may also request plotting).",
    )
    parser.add_argument(
        "--ft-as-feature",
        action="store_true",
        help="Alias for --ft-role embedding (keep tuning, export embeddings; skip FT plots/SHAP).",
    )
    parser.add_argument(
        "--ft-role",
        default=None,
        choices=["model", "embedding", "unsupervised_embedding"],
        help="How to use FT: model (default), embedding (export pooling embeddings), or unsupervised_embedding.",
    )
    parser.add_argument(
        "--ft-feature-prefix",
        default="ft_feat",
        help="Prefix used for generated FT features (columns: pred_<prefix>_0.. or pred_<prefix>).",
    )
    parser.add_argument(
        "--reuse-best-params",
        action="store_true",
        help="Skip Optuna and reuse best_params saved in Results/versions or bestparams CSV when available.",
    )
    return parser.parse_args()


def plot_curves_for_model(
    model: Any,
    trained_keys: List[str],
    cfg: Dict[str, Any],
    *,
    dedupe_preserve_order,
    plot_model_labels: Mapping[str, Tuple[str, str]],
    parse_model_pairs,
) -> None:
    plot_cfg = cfg.get("plot", {})
    plot_enabled = bool(plot_cfg.get("enable", False))
    if not plot_enabled:
        return

    n_bins = int(plot_cfg.get("n_bins", 10))
    oneway_enabled = plot_cfg.get("oneway", True)

    available_models = dedupe_preserve_order(
        [m for m in trained_keys if m in plot_model_labels]
    )

    lift_models = plot_cfg.get("lift_models")
    if lift_models is None:
        lift_models = available_models
    lift_models = dedupe_preserve_order(
        [m for m in lift_models if m in available_models]
    )

    if oneway_enabled:
        oneway_pred = bool(plot_cfg.get("oneway_pred", False))
        oneway_pred_models = plot_cfg.get("oneway_pred_models")
        pred_plotted = False
        if oneway_pred:
            if oneway_pred_models is None:
                oneway_pred_models = lift_models or available_models
            oneway_pred_models = dedupe_preserve_order(
                [m for m in oneway_pred_models if m in available_models]
            )
            for model_key in oneway_pred_models:
                label, pred_nme = plot_model_labels[model_key]
                if pred_nme not in model.train_data.columns:
                    print(
                        f"[Oneway] Missing prediction column '{pred_nme}'; skip.",
                        flush=True,
                    )
                    continue
                model.plot_oneway(
                    n_bins=n_bins,
                    pred_col=pred_nme,
                    pred_label=label,
                    plot_subdir="oneway/post",
                )
                pred_plotted = True
        if not oneway_pred or not pred_plotted:
            model.plot_oneway(n_bins=n_bins, plot_subdir="oneway/post")

    if not available_models:
        return

    for model_key in lift_models:
        label, pred_nme = plot_model_labels[model_key]
        model.plot_lift(model_label=label, pred_nme=pred_nme, n_bins=n_bins)

    if not plot_cfg.get("double_lift", True) or len(available_models) < 2:
        return

    raw_pairs = plot_cfg.get("double_lift_pairs")
    if raw_pairs:
        pairs = [
            (a, b)
            for a, b in parse_model_pairs(raw_pairs)
            if a in available_models and b in available_models and a != b
        ]
    else:
        pairs = [(a, b) for i, a in enumerate(available_models)
                 for b in available_models[i + 1:]]

    for first, second in pairs:
        model.plot_dlift([first, second], n_bins=n_bins)


def plot_loss_curve_for_trainer(
    model_name: str,
    trainer: Any,
    *,
    plot_loss_curve_common=None,
) -> None:
    model_obj = getattr(trainer, "model", None)
    history = None
    if model_obj is not None:
        history = getattr(model_obj, "training_history", None)
    if not history:
        history = getattr(trainer, "training_history", None)
    if not history:
        return
    train_hist = list(history.get("train") or [])
    val_hist = list(history.get("val") or [])
    if not train_hist and not val_hist:
        return
    try:
        plot_dir = trainer.output.plot_path(
            f"{model_name}/loss/loss_{model_name}_{trainer.model_name_prefix}.png"
        )
    except Exception:
        default_dir = Path("plot") / model_name / "loss"
        default_dir.mkdir(parents=True, exist_ok=True)
        plot_dir = str(
            default_dir / f"loss_{model_name}_{trainer.model_name_prefix}.png")
    if plot_loss_curve_common is not None:
        plot_loss_curve_common(
            history=history,
            title=f"{trainer.model_name_prefix} Loss Curve ({model_name})",
            save_path=plot_dir,
            show=False,
        )
    else:
        fig, ax = plt.subplots(figsize=(8, 4))
        if train_hist:
            ax.plot(range(1, len(train_hist) + 1),
                    train_hist, label="Train Loss", color="tab:blue")
        if val_hist:
            ax.plot(range(1, len(val_hist) + 1),
                    val_hist, label="Validation Loss", color="tab:orange")
        ax.set_xlabel("Epoch")
        ax.set_ylabel("Weighted Loss")
        ax.set_title(
            f"{trainer.model_name_prefix} Loss Curve ({model_name})")
        ax.grid(True, linestyle="--", alpha=0.3)
        ax.legend()
        plt.tight_layout()
        plt.savefig(plot_dir, dpi=300)
        plt.close(fig)
    print(
        f"[Plot] Saved loss curve for {model_name}/{trainer.label} -> {plot_dir}")


__all__ = [
    "parse_bayesopt_args",
    "plot_curves_for_model",
    "plot_loss_curve_for_trainer",
]



"""
CLI entry point generated from BayesOpt_AutoPricing.ipynb so the workflow can
run non-interactively (e.g., via torchrun).

Example:
    python -m torch.distributed.run --standalone --nproc_per_node=2 \\
        ins_pricing/cli/BayesOpt_entry.py \\
        --config-json examples/config_template.json \\
        --model-keys ft --max-evals 50 --use-ft-ddp
"""

from __future__ import annotations

from pathlib import Path
import importlib.util
import sys


def _ensure_repo_root() -> None:
    if __package__ not in {None, ""}:
        return
    if importlib.util.find_spec("ins_pricing") is not None:
        return
    bootstrap_path = Path(__file__).resolve().parents[1] / "utils" / "bootstrap.py"
    spec = importlib.util.spec_from_file_location("ins_pricing.cli.utils.bootstrap", bootstrap_path)
    if spec is None or spec.loader is None:
        return
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    module.ensure_repo_root()


_ensure_repo_root()

import argparse
from typing import Any, Dict, List, Optional

# Use unified import resolver to eliminate nested try/except chains
from ins_pricing.cli.utils import bayesopt_runner_reporting as reporting_utils
from ins_pricing.cli.utils.bayesopt_runner_ui import (
    parse_bayesopt_args,
    plot_curves_for_model,
    plot_loss_curve_for_trainer,
)
from ins_pricing.cli.utils.import_resolver import resolve_imports, setup_sys_path
from ins_pricing.cli.utils.evaluation_context import TrainingContext
from ins_pricing.modelling.bayesopt.runtime import (
    BayesOptRunnerDeps,
    BayesOptRunnerHooks,
    run_bayesopt_entry_training,
)

# Resolve all imports from a single location
setup_sys_path()
_imports = resolve_imports()

ropt = _imports.bayesopt
PLOT_MODEL_LABELS = _imports.PLOT_MODEL_LABELS
PYTORCH_TRAINERS = _imports.PYTORCH_TRAINERS
build_model_names = _imports.build_model_names
dedupe_preserve_order = _imports.dedupe_preserve_order
load_dataset = _imports.load_dataset
parse_model_pairs = _imports.parse_model_pairs
resolve_data_path = _imports.resolve_data_path
resolve_path = _imports.resolve_path
fingerprint_file = _imports.fingerprint_file
coerce_dataset_types = _imports.coerce_dataset_types
split_train_test = _imports.split_train_test

add_config_json_arg = _imports.add_config_json_arg
add_output_dir_arg = _imports.add_output_dir_arg
resolve_and_load_config = _imports.resolve_and_load_config
resolve_data_config = _imports.resolve_data_config
resolve_report_config = _imports.resolve_report_config
resolve_split_config = _imports.resolve_split_config
resolve_runtime_config = _imports.resolve_runtime_config
resolve_output_dirs = _imports.resolve_output_dirs

configure_run_logging = _imports.configure_run_logging
plot_loss_curve_common = _imports.plot_loss_curve

reporting_utils.configure_reporting_dependencies(
    plot_model_labels=PLOT_MODEL_LABELS,
    bootstrap_ci=_imports.bootstrap_ci,
    calibrate_predictions=_imports.calibrate_predictions,
    metrics_report=_imports.metrics_report,
    select_threshold=_imports.select_threshold,
    model_artifact_cls=_imports.ModelArtifact,
    model_registry_cls=_imports.ModelRegistry,
    drift_psi_report=_imports.drift_psi_report,
    group_metrics=_imports.group_metrics,
    report_payload_cls=_imports.ReportPayload,
    write_report=_imports.write_report,
)


def _parse_args() -> argparse.Namespace:
    return parse_bayesopt_args(
        add_config_json_arg=add_config_json_arg,
        add_output_dir_arg=add_output_dir_arg,
    )


def _plot_curves_for_model(model: ropt.BayesOptModel, trained_keys: List[str], cfg: Dict) -> None:
    plot_curves_for_model(
        model,
        trained_keys,
        cfg,
        dedupe_preserve_order=dedupe_preserve_order,
        plot_model_labels=PLOT_MODEL_LABELS,
        parse_model_pairs=parse_model_pairs,
    )


def _plot_loss_curve_for_trainer(model_name: str, trainer) -> None:
    plot_loss_curve_for_trainer(
        model_name,
        trainer,
        plot_loss_curve_common=plot_loss_curve_common,
    )


def _compute_psi_report(
    model: ropt.BayesOptModel,
    *,
    features: Optional[List[str]],
    bins: int,
    strategy: str,
) -> Any:
    return reporting_utils.compute_psi_report(
        model,
        features=features,
        bins=bins,
        strategy=strategy,
    )


def _evaluate_and_report(
    model: ropt.BayesOptModel,
    *,
    model_name: str,
    model_key: str,
    cfg: Dict[str, Any],
    data_path: Path,
    data_fingerprint: Dict[str, Any],
    report_output_dir: Optional[str],
    report_group_cols: Optional[List[str]],
    report_time_col: Optional[str],
    report_time_freq: str,
    report_time_ascending: bool,
    psi_report_df: Any,
    calibration_cfg: Dict[str, Any],
    threshold_cfg: Dict[str, Any],
    bootstrap_cfg: Dict[str, Any],
    register_model: bool,
    registry_path: Optional[str],
    registry_tags: Dict[str, Any],
    registry_status: str,
    run_id: str,
    config_sha: str,
) -> None:
    reporting_utils.evaluate_and_report(
        model,
        model_name=model_name,
        model_key=model_key,
        cfg=cfg,
        data_path=data_path,
        data_fingerprint=data_fingerprint,
        report_output_dir=report_output_dir,
        report_group_cols=report_group_cols,
        report_time_col=report_time_col,
        report_time_freq=report_time_freq,
        report_time_ascending=report_time_ascending,
        psi_report_df=psi_report_df,
        calibration_cfg=calibration_cfg,
        threshold_cfg=threshold_cfg,
        bootstrap_cfg=bootstrap_cfg,
        register_model=register_model,
        registry_path=registry_path,
        registry_tags=registry_tags,
        registry_status=registry_status,
        run_id=run_id,
        config_sha=config_sha,
    )


def _evaluate_with_context(
    model: ropt.BayesOptModel,
    ctx: Any,
) -> None:
    reporting_utils.evaluate_with_context(model, ctx)

def train_from_config(args: argparse.Namespace) -> None:
    deps = BayesOptRunnerDeps(
        ropt=ropt,
        pytorch_trainers=PYTORCH_TRAINERS,
        build_model_names=build_model_names,
        dedupe_preserve_order=dedupe_preserve_order,
        load_dataset=load_dataset,
        resolve_data_path=resolve_data_path,
        resolve_path=resolve_path,
        fingerprint_file=fingerprint_file,
        coerce_dataset_types=coerce_dataset_types,
        split_train_test=split_train_test,
        resolve_and_load_config=resolve_and_load_config,
        resolve_data_config=resolve_data_config,
        resolve_report_config=resolve_report_config,
        resolve_split_config=resolve_split_config,
        resolve_runtime_config=resolve_runtime_config,
        resolve_output_dirs=resolve_output_dirs,
    )
    hooks = BayesOptRunnerHooks(
        plot_loss_curve_for_trainer=_plot_loss_curve_for_trainer,
        compute_psi_report=_compute_psi_report,
        evaluate_and_report=_evaluate_and_report,
        plot_curves_for_model=_plot_curves_for_model,
    )
    run_bayesopt_entry_training(
        args,
        script_dir=Path(__file__).resolve().parents[1],
        deps=deps,
        hooks=hooks,
        training_context_from_env=TrainingContext.from_env,
    )

def main() -> None:
    if configure_run_logging:
        configure_run_logging(prefix="bayesopt_entry")
    args = _parse_args()
    train_from_config(args)


if __name__ == "__main__":
    main()
